# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np
import pytest

import pyttb as ttb

DEBUG_tests = True


@pytest.fixture()
def sample_tensor_2way():
    data = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    shape = (2, 3)
    params = {"data": data, "shape": shape}
    tensorInstance = ttb.tensor().from_data(data, shape)
    return params, tensorInstance


@pytest.fixture()
def sample_tensor_3way():
    data = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0])
    shape = (2, 3, 2)
    params = {"data": np.reshape(data, np.array(shape), order="F"), "shape": shape}
    tensorInstance = ttb.tensor().from_data(data, shape)
    return params, tensorInstance


@pytest.fixture()
def sample_tensor_4way():
    data = np.arange(1, 82)
    shape = (3, 3, 3, 3)
    params = {"data": np.reshape(data, np.array(shape), order="F"), "shape": shape}
    tensorInstance = ttb.tensor().from_data(data, shape)
    return params, tensorInstance


@pytest.mark.indevelopment
def test_tensor_initialization_empty():
    empty = np.array([])

    # No args
    tensorInstance = ttb.tensor()
    assert (tensorInstance.data == empty).all()
    assert tensorInstance.shape == ()


@pytest.mark.indevelopment
def test_tensor_initialization_from_data(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way
    assert (tensorInstance.data == params["data"]).all()
    assert tensorInstance.shape == params["shape"]

    with pytest.raises(AssertionError) as excinfo:
        a = ttb.tensor.from_data(params["data"], ())
    assert "Empty tensor cannot contain any elements" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        a = ttb.tensor.from_data(params["data"], (2, 4))
    assert "TTB:WrongSize, Size of data does not match specified size of tensor" in str(
        excinfo
    )

    with pytest.raises(AssertionError) as excinfo:
        a = ttb.tensor.from_data(params["data"], np.array([2, 3]))
    assert "Second argument must be a tuple." in str(excinfo)

    # TODO how else to break this logical statement?
    data = np.array([["a", 2, 3], [4, 5, 6]])
    with pytest.raises(AssertionError) as excinfo:
        a = ttb.tensor.from_data(data, (2, 3))
    assert "First argument must be a multidimensional array." in str(excinfo)

    # 1D tensors
    # no shape spaecified
    tensorInstance1 = ttb.tensor.from_data(np.array([1, 2, 3]))
    if DEBUG_tests:
        print("\ntensorInstance1:")
        print(tensorInstance1)
    data = np.array([1, 2, 3])
    assert tensorInstance1.data.shape == data.shape
    assert (tensorInstance1.data == data).all()

    # shape is 1 x 3
    tensorInstance1 = ttb.tensor.from_data(np.array([1, 2, 3]), (1, 3))
    if DEBUG_tests:
        print("\ntensorInstance1:")
        print(tensorInstance1)
    data = np.array([[1, 2, 3]])
    assert tensorInstance1.data.shape == data.shape
    assert (tensorInstance1.data == data).all()

    # shape is 3 x 1
    tensorInstance1 = ttb.tensor.from_data(np.array([1, 2, 3]), (3, 1))
    if DEBUG_tests:
        print("\ntensorInstance1:")
        print(tensorInstance1)
    data = np.array([[1], [2], [3]])
    assert tensorInstance1.data.shape == data.shape
    assert (tensorInstance1.data == data).all()


@pytest.mark.indevelopment
def test_tensor_initialization_from_tensor_type(sample_tensor_2way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way
    (_, tensorInstance4) = sample_tensor_4way

    # Copy Constructor
    tensorCopy = ttb.tensor.from_tensor_type(tensorInstance)
    assert (tensorCopy.data == params["data"]).all()
    assert tensorCopy.shape == params["shape"]

    subs = np.array([[0, 0], [0, 1], [0, 2], [1, 0]])
    vals = np.array([[1], [2], [3], [4]])
    shape = (2, 3)
    data = np.array([[1, 2, 3], [4, 0, 0]])
    a = ttb.sptensor.from_data(subs, vals, shape)

    # Sptensor
    b = ttb.tensor.from_tensor_type(a)
    assert (b.data == data).all()
    assert b.shape == shape

    # tenmat
    tenmatInstance = ttb.tenmat.from_tensor_type(tensorInstance, np.array([0]))
    tensorTenmatInstance = ttb.tensor.from_tensor_type(tenmatInstance)
    assert tensorInstance.isequal(tensorTenmatInstance)

    # 1D 1-element tenmat
    tensorInstance1 = ttb.tensor.from_data(np.array([3]))
    tenmatInstance1 = ttb.tenmat.from_tensor_type(tensorInstance1, np.array([0]))
    tensorTenmatInstance1 = ttb.tensor.from_tensor_type(tenmatInstance1)
    assert tensorInstance1.isequal(tensorTenmatInstance1)

    # 4D tenmat
    tenmatInstance4 = ttb.tenmat.from_tensor_type(tensorInstance4, np.array([3, 0]))
    tensorTenmatInstance4 = ttb.tensor.from_tensor_type(tenmatInstance4)
    assert tensorInstance4.isequal(tensorTenmatInstance4)

    # Non-tensor type
    with pytest.raises(ValueError):
        ttb.tensor.from_tensor_type(1)


@pytest.mark.indevelopment
def test_tensor_initialization_from_function():
    def function_handle(x):
        return np.array([[1, 2, 3], [4, 5, 6]])

    shape = (2, 3)
    data = np.array([[1, 2, 3], [4, 5, 6]])

    a = ttb.tensor.from_function(function_handle, shape)
    assert (a.data == data).all()
    assert a.shape == shape

    with pytest.raises(AssertionError) as excinfo:
        ttb.tensor.from_function(function_handle, [2, 3])
    assert "TTB:BadInput, Shape must be a tuple" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor_find(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way
    subs, vals = tensorInstance.find()
    if DEBUG_tests:
        print("\nsubs:")
        print(subs)
        print("\nvals:")
        print(vals)
    a = ttb.tensor.from_tensor_type(
        ttb.sptensor.from_data(subs, vals, tensorInstance.shape)
    )
    assert (a.data == tensorInstance.data).all()
    assert a.shape == tensorInstance.shape

    (params, tensorInstance) = sample_tensor_3way
    subs, vals = tensorInstance.find()
    if DEBUG_tests:
        print("\nsubs:")
        print(subs)
        print("\nvals:")
        print(vals)
    a = ttb.tensor.from_tensor_type(
        ttb.sptensor.from_data(subs, vals, tensorInstance.shape)
    )
    assert (a.data == tensorInstance.data).all()
    assert a.shape == tensorInstance.shape

    (params, tensorInstance) = sample_tensor_4way
    subs, vals = tensorInstance.find()
    if DEBUG_tests:
        print("\nsubs:")
        print(subs)
        print("\nvals:")
        print(vals)
    a = ttb.tensor.from_tensor_type(
        ttb.sptensor.from_data(subs, vals, tensorInstance.shape)
    )
    assert (a.data == tensorInstance.data).all()
    assert a.shape == tensorInstance.shape


@pytest.mark.indevelopment
def test_tensor_ndims(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way
    assert tensorInstance.ndims == len(params["shape"])

    (params, tensorInstance) = sample_tensor_3way
    assert tensorInstance.ndims == len(params["shape"])

    (params, tensorInstance) = sample_tensor_4way
    assert tensorInstance.ndims == len(params["shape"])

    # Empty tensor has zero dimensions
    assert ttb.tensor.from_data(np.array([])) == 0


@pytest.mark.indevelopment
def test_tensor__setitem__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Subtensor assign with constant
    dataCopy = params["data"].copy()
    dataCopy[1, 1] = 0.0
    tensorInstance[1, 1] = 0.0
    assert (tensorInstance.data == dataCopy).all()
    dataGrowth = np.zeros((5, 5))
    dataGrowth[0:2, 0:3] = dataCopy
    tensorInstance[4, 4] = 0.0
    assert (tensorInstance.data == dataGrowth).all()
    tensorInstance[0:2, 0:2] = 99.0
    dataGrowth[0:2, 0:2] = 99.0
    assert (tensorInstance.data == dataGrowth).all()

    # Subtensor assign with np array
    tensorInstance[0:2, 0:3] = dataCopy
    dataGrowth[0:2, 0:3] = dataCopy
    assert (tensorInstance.data == dataGrowth).all()

    # Subtensor assign with tensor
    tensorInstance[:, :] = tensorInstance
    assert (tensorInstance.data == dataGrowth).all()

    # Subscripts with constant
    tensorInstance[np.array([[1, 1]])] = 13.0
    dataGrowth[1, 1] = 13.0
    assert (tensorInstance.data == dataGrowth).all()

    # Subscripts with array
    tensorVector = ttb.tensor.from_data(np.array([0, 0, 0, 0]))
    tensorVector[np.array([0, 1, 2])] = np.array([3, 4, 5])
    assert (tensorVector.data == np.array([3, 4, 5, 0])).all()

    # Subscripts with constant
    tensorInstance[np.array([[1, 1], [1, 2]])] = 13.0
    dataGrowth[([1, 1], [1, 2])] = 13.0
    assert (tensorInstance.data == dataGrowth).all()

    # Linear Index with constant
    tensorInstance[np.array([0])] = 13.0
    dataGrowth[np.unravel_index([0], dataGrowth.shape, "F")] = 13.0
    assert (tensorInstance.data == dataGrowth).all()

    # Linear Index with constant
    tensorInstance[np.array([0, 3, 4])] = 13.0
    dataGrowth[np.unravel_index([0, 3, 4], dataGrowth.shape, "F")] = 13
    assert (tensorInstance.data == dataGrowth).all()

    # Test Empty Tensor Set Item, subtensor
    emptyTensor = ttb.tensor.from_data(np.array([]))
    emptyTensor[0, 0, 0] = 0
    assert (emptyTensor.data == np.array([[[0]]])).all()
    assert emptyTensor.shape == (1, 1, 1)

    # Test Empty Tensor Set Item, subscripts
    emptyTensor = ttb.tensor.from_data(np.array([]))
    emptyTensor[np.array([0, 0, 0])] = 0
    assert (emptyTensor.data == np.array([[[0]]])).all()
    assert emptyTensor.shape == (1, 1, 1)

    # Linear Index with constant, index out of bounds
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance[np.array([0, 3, 99])] = 13.0
    assert "TTB:BadIndex In assignment X[I] = Y, a tensor X cannot be resized" in str(
        excinfo
    )

    # Attempting to set some other way
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance[0, "a", 5] = 13.0
    assert "Invalid use of tensor setitem" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor__getitem__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way
    # Case 1 single element
    assert tensorInstance[0, 0] == params["data"][0, 0]
    # Case 1 Subtensor
    assert (tensorInstance[:, :] == tensorInstance).data.all()
    three_way_data = np.random.random((2, 3, 4))
    two_slices = (slice(None, None, None), 0, slice(None, None, None))
    assert (
        ttb.tensor.from_data(three_way_data)[two_slices].double()
        == three_way_data[two_slices]
    ).all()
    # Case 1 Subtensor
    assert (
        tensorInstance[np.array([0, 1]), :].data == tensorInstance.data[[0, 1], :]
    ).all()
    # Case 1 Subtensor
    assert (tensorInstance[0, :].data == tensorInstance.data[0, :]).all()
    assert (tensorInstance[:, 0].data == tensorInstance.data[:, 0]).all()

    # Case 2a:
    assert tensorInstance[np.array([0, 0]), "extract"] == params["data"][0, 0]
    assert (
        tensorInstance[np.array([[0, 0], [1, 1]]), "extract"]
        == params["data"][([0, 1], [0, 1])]
    ).all()

    # Case 2b: Linear Indexing
    assert tensorInstance[np.array([0])] == params["data"][0, 0]
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance[np.array([0]), np.array([0]), np.array([0])]
    assert "Linear indexing requires single input array" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor_logical_and(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor And
    assert (
        tensorInstance.logical_and(tensorInstance).data == np.ones((params["shape"]))
    ).all()

    # Non-zero And
    assert (tensorInstance.logical_and(1).data == np.ones((params["shape"]))).all()

    # Zero And
    assert (tensorInstance.logical_and(0).data == np.zeros((params["shape"]))).all()


@pytest.mark.indevelopment
def test_tensor__eq__(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor tensor equality
    assert ((tensorInstance == tensorInstance).data).all()

    # Tensor scalar equality, not equal
    assert not ((tensorInstance == 7).data).any()

    # Tensor scalar equality, is equal
    data = np.zeros(params["data"].shape)
    data[0, 0] = 1
    assert ((tensorInstance == 1).data == data).all()

    (params3, tensorInstance3) = sample_tensor_3way

    # Tensor tensor equality
    assert ((tensorInstance3 == tensorInstance3).data).all()

    (params4, tensorInstance4) = sample_tensor_4way

    # Tensor tensor equality
    assert ((tensorInstance4 == tensorInstance4).data).all()


@pytest.mark.indevelopment
def test_tensor__ne__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor tensor equality
    assert not ((tensorInstance != tensorInstance).data).any()

    # Tensor scalar equality, not equal
    assert ((tensorInstance != 7).data).all()

    # Tensor scalar equality, is equal
    data = np.zeros(params["data"].shape)
    data[0, 0] = 1
    assert not ((tensorInstance != 1).data == data).any()


@pytest.mark.indevelopment
def test_tensor_full(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params2, tensorInstance2) = sample_tensor_2way
    if DEBUG_tests:
        print("\nparam2['data']:")
        print(params2["data"])
        print("\ntensorInstace2.data:")
        print(tensorInstance2.data)
        print("\ntensorInstace2.full():")
        print(tensorInstance2.full())
    assert (tensorInstance2.full().data == params2["data"]).all()

    (params3, tensorInstance3) = sample_tensor_3way
    if DEBUG_tests:
        print("\nparam3['data']:")
        print(params3["data"])
        print("\ntensorInstace3.data:")
        print(tensorInstance3.data)
        print("\ntensorInstace3.full():")
        print(tensorInstance3.full())
    assert (tensorInstance3.full().data == params3["data"]).all()

    (params4, tensorInstance4) = sample_tensor_4way
    if DEBUG_tests:
        print("\nparam4['data']:")
        print(params4["data"])
        print("\ntensorInstace4.data:")
        print(tensorInstance4.data)
        print("\ntensorInstace4.full():")
        print(tensorInstance4.full())
    assert (tensorInstance4.full().data == params4["data"]).all()


@pytest.mark.indevelopment
def test_tensor__ge__(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert ((tensorInstance >= tensorInstance).data).all()
    assert ((tensorInstance >= tensorSmaller).data).all()
    assert not ((tensorInstance >= tensorLarger).data).any()

    (params, tensorInstance) = sample_tensor_3way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert ((tensorInstance >= tensorInstance).data).all()
    assert ((tensorInstance >= tensorSmaller).data).all()
    assert not ((tensorInstance >= tensorLarger).data).any()

    (params, tensorInstance) = sample_tensor_4way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert ((tensorInstance >= tensorInstance).data).all()
    assert ((tensorInstance >= tensorSmaller).data).all()
    assert not ((tensorInstance >= tensorLarger).data).any()


@pytest.mark.indevelopment
def test_tensor__gt__(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert not ((tensorInstance > tensorInstance).data).any()
    assert ((tensorInstance > tensorSmaller).data).all()
    assert not ((tensorInstance > tensorLarger).data).any()

    (params, tensorInstance) = sample_tensor_3way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert not ((tensorInstance > tensorInstance).data).any()
    assert ((tensorInstance > tensorSmaller).data).all()
    assert not ((tensorInstance > tensorLarger).data).any()

    (params, tensorInstance) = sample_tensor_4way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert not ((tensorInstance > tensorInstance).data).any()
    assert ((tensorInstance > tensorSmaller).data).all()
    assert not ((tensorInstance > tensorLarger).data).any()


@pytest.mark.indevelopment
def test_tensor__le__(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert ((tensorInstance <= tensorInstance).data).all()
    assert not ((tensorInstance <= tensorSmaller).data).any()
    assert ((tensorInstance <= tensorLarger).data).all()

    (params, tensorInstance) = sample_tensor_3way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert ((tensorInstance <= tensorInstance).data).all()
    assert not ((tensorInstance <= tensorSmaller).data).any()
    assert ((tensorInstance <= tensorLarger).data).all()

    (params, tensorInstance) = sample_tensor_4way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert ((tensorInstance <= tensorInstance).data).all()
    assert not ((tensorInstance <= tensorSmaller).data).any()
    assert ((tensorInstance <= tensorLarger).data).all()


@pytest.mark.indevelopment
def test_tensor__lt__(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert not ((tensorInstance < tensorInstance).data).any()
    assert not ((tensorInstance < tensorSmaller).data).any()
    assert ((tensorInstance < tensorLarger).data).all()

    (params, tensorInstance) = sample_tensor_3way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert not ((tensorInstance < tensorInstance).data).any()
    assert not ((tensorInstance < tensorSmaller).data).any()
    assert ((tensorInstance < tensorLarger).data).all()

    (params, tensorInstance) = sample_tensor_4way

    tensorLarger = ttb.tensor.from_data(params["data"] + 1)
    tensorSmaller = ttb.tensor.from_data(params["data"] - 1)

    assert not ((tensorInstance < tensorInstance).data).any()
    assert not ((tensorInstance < tensorSmaller).data).any()
    assert ((tensorInstance < tensorLarger).data).all()


@pytest.mark.indevelopment
def test_tensor_norm(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    # 2-way tensor
    (params2, tensorInstance2) = sample_tensor_2way
    if DEBUG_tests:
        print("\ntensorInstace2.norm():")
        print(tensorInstance2.norm())
    assert tensorInstance2.norm() == np.linalg.norm(params2["data"].ravel())

    # 3-way tensor
    (params3, tensorInstance3) = sample_tensor_3way
    if DEBUG_tests:
        print("\ntensorInstace3.norm():")
        print(tensorInstance3.norm())
    assert tensorInstance3.norm() == np.linalg.norm(params3["data"].ravel())

    # 4-way tensor
    (params4, tensorInstance4) = sample_tensor_4way
    if DEBUG_tests:
        print("\ntensorInstace4.norm():")
        print(tensorInstance4.norm())
    assert tensorInstance4.norm() == np.linalg.norm(params4["data"].ravel())


@pytest.mark.indevelopment
def test_tensor_logical_not(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    assert (tensorInstance.logical_not().data == np.logical_not(params["data"])).all()


@pytest.mark.indevelopment
def test_tensor_logical_or(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor Or
    assert (
        tensorInstance.logical_or(tensorInstance).data == np.ones((params["shape"]))
    ).all()

    # Non-zero Or
    assert (tensorInstance.logical_or(1).data == np.ones((params["shape"]))).all()

    # Zero Or
    assert (tensorInstance.logical_or(0).data == np.ones((params["shape"]))).all()


@pytest.mark.indevelopment
def test_tensor_logical_xor(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor Or
    assert (
        tensorInstance.logical_xor(tensorInstance).data == np.zeros((params["shape"]))
    ).all()

    # Non-zero Or
    assert (tensorInstance.logical_xor(1).data == np.zeros((params["shape"]))).all()

    # Zero Or
    assert (tensorInstance.logical_xor(0).data == np.ones((params["shape"]))).all()


@pytest.mark.indevelopment
def test_tensor__add__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor + Tensor
    assert ((tensorInstance + tensorInstance).data == 2 * (params["data"])).all()

    # Tensor + scalar
    assert ((tensorInstance + 1).data == 1 + (params["data"])).all()

    # scalar + Tensor
    assert ((1 + tensorInstance).data == 1 + (params["data"])).all()


@pytest.mark.indevelopment
def test_tensor__sub__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor - Tensor
    assert ((tensorInstance - tensorInstance).data == 0 * (params["data"])).all()

    # Tensor - scalar
    assert ((tensorInstance - 1).data == (params["data"] - 1)).all()


@pytest.mark.indevelopment
def test_tensor__pow__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor** Tensor
    assert (
        (tensorInstance**tensorInstance).data == (params["data"] ** params["data"])
    ).all()
    # Tensor**Scalar
    assert ((tensorInstance**2).data == (params["data"] ** 2)).all()


@pytest.mark.indevelopment
def test_tensor__mul__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor* Tensor
    assert (
        (tensorInstance * tensorInstance).data == (params["data"] * params["data"])
    ).all()
    # Tensor*Scalar
    assert ((tensorInstance * 2).data == (params["data"] * 2)).all()
    # Tensor * Sptensor
    assert (
        (tensorInstance * ttb.sptensor.from_tensor_type(tensorInstance)).data
        == (params["data"] * params["data"])
    ).all()

    # TODO tensor * ktensor


@pytest.mark.indevelopment
def test_tensor__rmul__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Scalar * Tensor, only resolves when left object doesn't have appropriate __mul__
    assert ((2 * tensorInstance).data == (params["data"] * 2)).all()


@pytest.mark.indevelopment
def test_tensor__pos__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # +Tensor yields no change
    assert ((+tensorInstance).data == params["data"]).all()


@pytest.mark.indevelopment
def test_tensor__neg__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # -Tensor yields negated copy of tensor
    assert ((-tensorInstance).data == -1 * params["data"]).all()
    # Original tensor should remain unchanged
    assert ((tensorInstance).data == params["data"]).all()


@pytest.mark.indevelopment
def test_tensor_double(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    assert (tensorInstance.double() == params["data"]).all()


@pytest.mark.indevelopment
def test_tensor_end(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    assert tensorInstance.end() == np.prod(params["shape"]) - 1
    assert tensorInstance.end(k=0) == params["shape"][0] - 1


@pytest.mark.indevelopment
def test_tensor_isequal(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way
    subs = []
    vals = []
    for j in range(3):
        for i in range(2):
            subs.append([i, j])
            vals.append([params["data"][i, j]])
    sptensorInstance = ttb.sptensor.from_data(
        np.array(subs), np.array(vals), params["shape"]
    )

    assert tensorInstance.isequal(tensorInstance)
    assert tensorInstance.isequal(sptensorInstance)

    # Tensor is not equal to scalar
    assert not tensorInstance.isequal(1)


@pytest.mark.indevelopment
def test_tensor__truediv__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Tensor / Tensor
    assert (
        (tensorInstance / tensorInstance).data == (params["data"] / params["data"])
    ).all()

    # Tensor / Sptensor
    assert (
        (tensorInstance / ttb.sptensor.from_tensor_type(tensorInstance)).data
        == (params["data"] / params["data"])
    ).all()

    # Tensor / Scalar
    assert ((tensorInstance / 2).data == (params["data"] / 2)).all()


@pytest.mark.indevelopment
def test_tensor__rtruediv__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Scalar / Tensor, only resolves when left object doesn't have appropriate __mul__
    assert ((2 / tensorInstance).data == (2 / params["data"])).all()


@pytest.mark.indevelopment
def test_tensor_nnz(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # NNZ for full tensor
    assert tensorInstance.nnz == 6

    # NNZ for tensor with at least one zero
    tensorInstance[0, 0] = 0
    assert tensorInstance.nnz == 5


@pytest.mark.indevelopment
def test_tensor_reshape(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params2, tensorInstance2) = sample_tensor_2way

    # Reshape with tuple
    tensorInstance2 = tensorInstance2.reshape((3, 2))
    if DEBUG_tests:
        print("\ntensorInstance2.reshape((3, 2)):")
        print(tensorInstance2.reshape((3, 2)))
    assert tensorInstance2.shape == (3, 2)
    data = np.array([[1.0, 5.0], [4.0, 3.0], [2.0, 6.0]])
    assert (tensorInstance2.data == data).all()

    with pytest.raises(AssertionError) as excinfo:
        tensorInstance2.reshape((3, 3))
    assert "Reshaping a tensor cannot change number of elements" in str(excinfo)

    (params3, tensorInstance3) = sample_tensor_3way
    tensorInstance3 = tensorInstance3.reshape((3, 2, 2))
    if DEBUG_tests:
        print("\ntensorInstance3.reshape((3, 2, 2)):")
        print(tensorInstance3)
    assert tensorInstance3.shape == (3, 2, 2)
    data3 = np.array(
        [
            [[1.0, 7.0], [4.0, 10.0]],
            [[2.0, 8.0], [5.0, 11.0]],
            [[3.0, 9.0], [6.0, 12.0]],
        ]
    )
    assert (tensorInstance3.data == data3).all()

    (params4, tensorInstance4) = sample_tensor_4way
    tensorInstance4 = tensorInstance4.reshape((1, 3, 3, 9))
    if DEBUG_tests:
        print("\ntensorInstance4.reshape((1, 3, 3, 9)):")
        print(tensorInstance4)
    assert tensorInstance4.shape == (1, 3, 3, 9)
    data4 = np.array(
        [
            [
                [
                    [1, 10, 19, 28, 37, 46, 55, 64, 73],
                    [4, 13, 22, 31, 40, 49, 58, 67, 76],
                    [7, 16, 25, 34, 43, 52, 61, 70, 79],
                ],
                [
                    [2, 11, 20, 29, 38, 47, 56, 65, 74],
                    [5, 14, 23, 32, 41, 50, 59, 68, 77],
                    [8, 17, 26, 35, 44, 53, 62, 71, 80],
                ],
                [
                    [3, 12, 21, 30, 39, 48, 57, 66, 75],
                    [6, 15, 24, 33, 42, 51, 60, 69, 78],
                    [9, 18, 27, 36, 45, 54, 63, 72, 81],
                ],
            ]
        ]
    )
    assert (tensorInstance4.data == data4).all()


@pytest.mark.indevelopment
def test_tensor_permute(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way

    # Permute rows and columns
    assert (
        tensorInstance.permute(np.array([1, 0])).data == np.transpose(params["data"])
    ).all()

    # len(order) != ndims
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance.permute(np.array([1, 0, 2]))
    assert "Invalid permutation order" in str(excinfo)

    # Try to permute order-1 tensor
    assert (
        ttb.tensor.from_data(np.array([1, 2, 3, 4])).permute(np.array([1])).data
        == np.array([1, 2, 3, 4])
    ).all()

    # Empty order
    assert (
        ttb.tensor.from_data(np.array([])).permute(np.array([])).data == np.array([])
    ).all()

    # 2-way
    (params2, tensorInstance2) = sample_tensor_2way
    tensorInstance2 = tensorInstance2.permute(np.array([1, 0]))
    if DEBUG_tests:
        print("\ntensorInstance2.permute(np.array([1, 0])):")
        print(tensorInstance2)
    assert tensorInstance2.shape == (3, 2)
    data2 = np.array([[1.0, 4.0], [2.0, 5.0], [3.0, 6.0]])
    assert (tensorInstance2.data == data2).all()

    # 3-way
    (params3, tensorInstance3) = sample_tensor_3way
    tensorInstance3 = tensorInstance3.permute(np.array([2, 1, 0]))
    if DEBUG_tests:
        print("\ntensorInstance3.permute(np.array([2, 1, 0])):")
        print(tensorInstance3)
    assert tensorInstance3.shape == (2, 3, 2)
    data3 = np.array(
        [[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], [[7.0, 8.0], [9.0, 10.0], [11.0, 12.0]]]
    )
    assert (tensorInstance3.data == data3).all()

    # 4-way
    (params4, tensorInstance4) = sample_tensor_4way
    tensorInstance4 = tensorInstance4.permute(np.array([3, 1, 2, 0]))
    if DEBUG_tests:
        print("\ntensorInstance4.permute(np.array([3, 1, 2, 0])):")
        print(tensorInstance4)
    assert tensorInstance4.shape == (3, 3, 3, 3)
    data4 = np.array(
        [
            [
                [[1, 2, 3], [10, 11, 12], [19, 20, 21]],
                [[4, 5, 6], [13, 14, 15], [22, 23, 24]],
                [[7, 8, 9], [16, 17, 18], [25, 26, 27]],
            ],
            [
                [[28, 29, 30], [37, 38, 39], [46, 47, 48]],
                [[31, 32, 33], [40, 41, 42], [49, 50, 51]],
                [[34, 35, 36], [43, 44, 45], [52, 53, 54]],
            ],
            [
                [[55, 56, 57], [64, 65, 66], [73, 74, 75]],
                [[58, 59, 60], [67, 68, 69], [76, 77, 78]],
                [[61, 62, 63], [70, 71, 72], [79, 80, 81]],
            ],
        ]
    )
    assert (tensorInstance4.data == data4).all()


@pytest.mark.indevelopment
def test_tensor_collapse(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params2, tensorInstance2) = sample_tensor_2way

    assert tensorInstance2.collapse() == 21
    assert tensorInstance2.collapse(fun=np.max) == 6

    (params3, tensorInstance3) = sample_tensor_3way

    assert tensorInstance3.collapse() == 78
    assert tensorInstance3.collapse(fun=np.max) == 12

    (params4, tensorInstance4) = sample_tensor_4way

    assert tensorInstance4.collapse() == 3321
    assert tensorInstance4.collapse(fun=np.max) == 81

    # single dimension collapse
    data = np.array([5, 7, 9])
    tensorCollapse = tensorInstance2.collapse(np.array([0]))
    assert (tensorCollapse.data == data).all()

    # single dimension collapse using max function
    datamax = np.array([4, 5, 6])
    tensorCollapseMax = tensorInstance2.collapse(np.array([0]), fun=np.max)
    assert (tensorCollapseMax.data == datamax).all()

    # multiple dimensions collapse
    data4 = np.array([[99, 342, 585], [126, 369, 612], [153, 396, 639]])
    tensorCollapse4 = tensorInstance4.collapse(np.array([0, 2]))
    assert (tensorCollapse4.data == data4).all()

    # multiple dimensions collapse
    data4max = np.array([[21, 48, 75], [24, 51, 78], [27, 54, 81]])
    tensorCollapse4Max = tensorInstance4.collapse(np.array([0, 2]), fun=np.max)
    assert (tensorCollapse4Max.data == data4max).all()

    # Empty tensor collapse
    empty_data = np.array([])
    empty_tensor = ttb.tensor.from_data(empty_data)
    assert np.all(empty_tensor.collapse() == empty_data)

    # Empty dims
    assert tensorInstance2.collapse(empty_data).isequal(tensorInstance2)


@pytest.mark.indevelopment
def test_tensor_contract(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params2, tensorInstance2) = sample_tensor_2way

    with pytest.raises(AssertionError) as excinfo:
        tensorInstance2.contract(0, 1)
    assert "Must contract along equally sized dimensions" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        tensorInstance2.contract(0, 0)
    assert "Must contract along two different dimensions" in str(excinfo)

    contractableTensor = ttb.tensor.from_data(
        np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]), (3, 3)
    )
    assert contractableTensor.contract(0, 1) == 15

    (params3, tensorInstance3) = sample_tensor_3way
    print("\ntensorInstance3.contract(0,2) = ")
    print(tensorInstance3.contract(0, 2))
    data3 = np.array([9, 13, 17])
    assert (tensorInstance3.contract(0, 2).data == data3).all()

    (params4, tensorInstance4) = sample_tensor_4way
    print("\ntensorInstance4.contract(0,1) = ")
    print(tensorInstance4.contract(0, 1))
    data4 = np.array([[15, 96, 177], [42, 123, 204], [69, 150, 231]])
    assert (tensorInstance4.contract(0, 1).data == data4).all()

    print("\ntensorInstance4.contract(1,3) = ")
    print(tensorInstance4.contract(1, 3))
    data4a = np.array([[93, 120, 147], [96, 123, 150], [99, 126, 153]])
    assert (tensorInstance4.contract(1, 3).data == data4a).all()


@pytest.mark.indevelopment
def test_tensor__repr__(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way
    # Test that we can capture each of these
    str(tensorInstance)

    str(ttb.tensor.from_data(np.array([1, 2, 3])))

    str(ttb.tensor.from_data(np.arange(0, 81).reshape(3, 3, 3, 3)))

    str(ttb.tensor())


@pytest.mark.indevelopment
def test_tensor_exp(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way
    assert (tensorInstance.exp().data == np.exp(params["data"])).all()


@pytest.mark.indevelopment
def test_tensor_innerprod(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params, tensorInstance) = sample_tensor_2way
    # Tensor innerproduct
    assert tensorInstance.innerprod(tensorInstance) == np.arange(1, 7).dot(
        np.arange(1, 7)
    )

    # Sptensor innerproduct
    assert tensorInstance.innerprod(
        ttb.sptensor.from_tensor_type(tensorInstance)
    ) == np.arange(1, 7).dot(np.arange(1, 7))

    # Wrong size innerproduct
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance.innerprod(ttb.tensor.from_data(np.ones((4, 4))))
    assert "Inner product must be between tensors of the same size" in str(excinfo)

    # Wrong class innerproduct
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance.innerprod(5)
    assert "Inner product between tensor and that class is not supported" in str(
        excinfo
    )

    # 2-way
    (params2, tensorInstance2) = sample_tensor_2way
    if DEBUG_tests:
        print(
            f"\ntensorInstance2.innerprod(tensorInstance2): {tensorInstance2.innerprod(tensorInstance2)}"
        )
    assert tensorInstance2.innerprod(tensorInstance2) == 91

    # 3-way
    (params3, tensorInstance3) = sample_tensor_3way
    if DEBUG_tests:
        print(
            f"\ntensorInstance3.innerprod(tensorInstance3): {tensorInstance3.innerprod(tensorInstance3)}"
        )
    assert tensorInstance3.innerprod(tensorInstance3) == 650

    # 4-way
    (params4, tensorInstance4) = sample_tensor_4way
    if DEBUG_tests:
        print(
            f"\ntensorInstance4.innerprod(tensorInstance4): {tensorInstance4.innerprod(tensorInstance4)}"
        )
    assert tensorInstance4.innerprod(tensorInstance4) == 180441


@pytest.mark.indevelopment
def test_tensor_mask(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way
    W = ttb.tensor.from_data(np.array([[0, 1, 0], [1, 0, 0]]))
    assert (tensorInstance.mask(W) == np.array([4, 2])).all()

    # Wrong shape mask
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance.mask(ttb.tensor.from_data(np.ones((11, 3))))
    assert "Mask cannot be bigger than the data tensor" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor_squeeze(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # No singleton dimensions
    assert (tensorInstance.squeeze().data == params["data"]).all()

    # All singleton dimensions
    squeeze_result = ttb.tensor.from_data(np.array([[[4]]])).squeeze()
    assert squeeze_result == 4
    assert np.isscalar(squeeze_result)

    # A singleton dimension
    assert (
        ttb.tensor.from_data(np.array([[1, 2, 3]])).squeeze().data
        == np.array([1, 2, 3])
    ).all()


@pytest.mark.indevelopment
def test_tensor_ttm(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params2, tensorInstance2) = sample_tensor_2way
    (params3, tensorInstance3) = sample_tensor_3way
    (params4, tensorInstance4) = sample_tensor_4way

    M2 = np.reshape(np.arange(1, 2 * 2 + 1), [2, 2], order="F")
    M3 = np.reshape(np.arange(1, 3 * 3 + 1), [3, 3], order="F")

    # 3-way single matrix
    T3 = tensorInstance3.ttm(M2, 0)
    assert isinstance(T3, ttb.tensor)
    assert T3.shape == (2, 3, 2)
    data3 = np.array([[[7, 31], [15, 39], [23, 47]], [[10, 46], [22, 58], [34, 70]]])
    assert (T3.data == data3).all()

    # 3-way single matrix, transposed
    T3 = tensorInstance3.ttm(M2, 0, transpose=True)
    assert isinstance(T3, ttb.tensor)
    assert T3.shape == (2, 3, 2)
    data3 = np.array([[[5, 23], [11, 29], [17, 35]], [[11, 53], [25, 67], [39, 81]]])
    assert (T3.data == data3).all()

    # 3-way, two matrices, negative dimension
    T3 = tensorInstance3.ttm([M2, M2], exclude_dims=1)
    assert isinstance(T3, ttb.tensor)
    assert T3.shape == (2, 3, 2)
    data3 = np.array(
        [[[100, 138], [132, 186], [164, 234]], [[148, 204], [196, 276], [244, 348]]]
    )
    assert (T3.data == data3).all()

    # 3-way, two matrices, explicit dimensions
    T3 = tensorInstance3.ttm([M2, M3], [2, 1])
    assert isinstance(T3, ttb.tensor)
    assert T3.shape == (2, 3, 2)
    data3 = np.array(
        [[[408, 576], [498, 702], [588, 828]], [[456, 648], [558, 792], [660, 936]]]
    )
    assert (T3.data == data3).all()

    # 3-way, 3 matrices, no dimensions specified
    T3 = tensorInstance3.ttm([M2, M3, M2])
    assert isinstance(T3, ttb.tensor)
    assert T3.shape == (2, 3, 2)
    data3 = np.array(
        [
            [[1776, 2520], [2172, 3078], [2568, 3636]],
            [[2640, 3744], [3228, 4572], [3816, 5400]],
        ]
    )
    assert (T3.data == data3).all()

    # 3-way, matrix must be np.ndarray
    Tmat = ttb.tenmat.from_data(M2, rdims=np.array([0]))
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance3.ttm(Tmat, 0)
    assert "matrix must be of type numpy.ndarray" in str(excinfo)

    # 3-way, dims must be in range [0,self.ndims]
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance3.ttm(M2, tensorInstance3.ndims + 1)
    assert "dims must contain values in [0,self.dims)" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor_ttt(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    M31 = ttb.tensor.from_data(
        np.reshape(np.arange(1, 2 * 3 * 4 + 1), [4, 3, 2], order="F")
    )
    M32 = ttb.tensor.from_data(
        np.reshape(np.arange(1, 2 * 3 * 4 + 1), [3, 4, 2], order="F")
    )

    # outer product of M31 and M32
    TTT1 = M31.ttt(M32)
    assert TTT1.shape == (4, 3, 2, 3, 4, 2)
    # choose two random 2-way slices
    data11 = np.array([1, 2, 3, 4])
    data12 = np.array([289, 306, 323, 340])
    data13 = np.array([504, 528, 552, 576])
    assert (TTT1[:, 0, 0, 0, 0, 0].data == data11).all()
    assert (TTT1[:, 1, 1, 1, 1, 1].data == data12).all()
    assert (TTT1[:, 2, 1, 2, 3, 1].data == data13).all()

    TTT1_with_dims = M31.ttt(
        M31, selfdims=np.array([0, 1, 2]), otherdims=np.array([0, 1, 2])
    )
    assert np.allclose(TTT1_with_dims, M31.innerprod(M31))

    # Negative tests
    with pytest.raises(AssertionError):
        invalid_tensor_type = []
        M31.ttt(invalid_tensor_type)

    with pytest.raises(AssertionError):
        M31.ttt(M31, selfdims=np.array([0, 1, 2]), otherdims=np.array([0, 2, 1]))

    M2 = ttb.tensor.from_data(np.reshape(np.arange(0, 2), [1, 2], order="F"))
    result = M2.ttt(M2, 0, 0)
    row_vector = M2.data
    column_vector = M2.data.transpose()
    assert np.allclose(result.data, row_vector * column_vector)


@pytest.mark.indevelopment
def test_tensor_ttv(sample_tensor_2way, sample_tensor_3way, sample_tensor_4way):
    (params2, tensorInstance2) = sample_tensor_2way
    (params3, tensorInstance3) = sample_tensor_3way
    (params4, tensorInstance4) = sample_tensor_4way

    # Wrong shape vector
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance2.ttv(np.array([np.array([1, 2]), np.array([1, 2])]))
    assert "Multiplicand is wrong size" in str(excinfo)

    # 2-way Multiply by single vector
    T2 = tensorInstance2.ttv(np.array([2, 2]), 0)
    assert isinstance(T2, ttb.tensor)
    assert T2.shape == (3,)
    assert (T2.data == np.array([10, 14, 18])).all()

    # 2-way Multiply by single vector (exclude dims)
    T2 = tensorInstance2.ttv(np.array([2, 2]), exclude_dims=1)
    assert isinstance(T2, ttb.tensor)
    assert T2.shape == (3,)
    assert (T2.data == np.array([10, 14, 18])).all()

    # Multiply by multiple vectors, infer dimensions
    assert tensorInstance2.ttv([np.array([2, 2]), np.array([1, 1, 1])]) == 42

    # Multiply by multiple vectors as list of numpy.ndarrays
    assert tensorInstance2.ttv([np.array([2, 2]), np.array([1, 1, 1])]) == 42

    # 3-way Multiply by single vector
    T3 = tensorInstance3.ttv(2 * np.ones((tensorInstance3.shape[0],)), 0)
    assert isinstance(T3, ttb.tensor)
    assert T3.shape == (tensorInstance3.shape[1], tensorInstance3.shape[2])
    assert (T3.data == np.array([[6, 30], [14, 38], [22, 46]])).all()

    # Multiply by multiple vectors, infer dimensions
    assert (
        tensorInstance3.ttv([np.array([2, 2]), np.array([1, 1, 1]), np.array([2, 2])])
        == 312
    )

    # 4-way Multiply by single vector
    T4 = tensorInstance4.ttv(2 * np.ones((tensorInstance4.shape[0],)), 0)
    assert isinstance(T4, ttb.tensor)
    assert T4.shape == (
        tensorInstance4.shape[1],
        tensorInstance4.shape[2],
        tensorInstance4.shape[3],
    )

    # 4-way Multiply by single vector (exclude dims)
    T4 = tensorInstance4.ttv(
        2 * np.ones((tensorInstance4.shape[0],)), exclude_dims=np.array([1, 2, 3])
    )
    assert isinstance(T4, ttb.tensor)
    assert T4.shape == (
        tensorInstance4.shape[1],
        tensorInstance4.shape[2],
        tensorInstance4.shape[3],
    )

    data4 = np.array(
        [
            [[12, 174, 336], [66, 228, 390], [120, 282, 444]],
            [[30, 192, 354], [84, 246, 408], [138, 300, 462]],
            [[48, 210, 372], [102, 264, 426], [156, 318, 480]],
        ]
    )
    assert (T4.data == data4).all()

    # Multiply by multiple vectors, infer dimensions
    assert (
        tensorInstance4.ttv(
            np.array(
                [
                    np.array([1, 1, 1]),
                    np.array([1, 1, 1]),
                    np.array([1, 1, 1]),
                    np.array([1, 1, 1]),
                ]
            )
        )
        == 3321
    )


@pytest.mark.indevelopment
def test_tensor_ttsv(sample_tensor_4way):
    # 3-way
    tensorInstance3 = ttb.tensor.from_data(np.ones((4, 4, 4)))
    vector3 = np.array([4, 3, 2, 1])
    assert tensorInstance3.ttsv(vector3, version=1) == 1000
    assert (
        tensorInstance3.ttsv(vector3, skip_dim=0, version=1) == 100 * np.ones((4,))
    ).all()
    assert (
        tensorInstance3.ttsv(vector3, skip_dim=1, version=1) == 10 * np.ones((4, 4))
    ).all()

    # Invalid dims
    with pytest.raises(ValueError) as excinfo:
        tensorInstance3.ttsv(vector3, skip_dim=-1)
    assert "Invalid modes in ttsv" in str(excinfo)

    # 4-way tensor
    (params4, tensorInstance4) = sample_tensor_4way
    T4ttsv = tensorInstance4.ttsv(np.array([1, 2, 3]), 2, version=1)
    data4_3 = np.array(
        [
            [[222, 276, 330], [240, 294, 348], [258, 312, 366]],
            [[228, 282, 336], [246, 300, 354], [264, 318, 372]],
            [[234, 288, 342], [252, 306, 360], [270, 324, 378]],
        ]
    )
    assert (T4ttsv.data == data4_3).all()

    # 5-way dense tensor
    shape = (3, 3, 3, 3, 3)
    T5 = ttb.tensor.from_data(np.arange(1, np.prod(shape) + 1), shape)
    T5ttsv = T5.ttsv(np.array([1, 2, 3]), 2, version=1)
    data5_3 = np.array(
        [
            [[5220, 5544, 5868], [5328, 5652, 5976], [5436, 5760, 6084]],
            [[5256, 5580, 5904], [5364, 5688, 6012], [5472, 5796, 6120]],
            [[5292, 5616, 5940], [5400, 5724, 6048], [5508, 5832, 6156]],
        ]
    )
    assert (T5ttsv.data == data5_3).all()

    # Test new algorithm, version=2

    # 3-way
    assert tensorInstance3.ttsv(vector3) == 1000
    assert (tensorInstance3.ttsv(vector3, 0) == 100 * np.ones((4,))).all()
    assert (tensorInstance3.ttsv(vector3, 1) == 10 * np.ones((4, 4))).all()

    # 4-way tensor
    T4ttsv2 = tensorInstance4.ttsv(np.array([1, 2, 3]), 2)
    assert (T4ttsv2.data == data4_3).all()

    # Incorrect version requested
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance4.ttsv(np.array([1, 2, 3]), 2, version=3)
    assert "Invalid value for version; should be None, 1, or 2" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor_issymmetric(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    assert tensorInstance.issymmetric() is False
    assert tensorInstance.issymmetric(version=1) is False

    symmetricData = np.array(
        [
            [[0.5, 0, 0.5, 0], [0, 0, 0, 0], [0.5, 0, 0, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 2.5, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
            [[0.5, 0, 0, 0], [0, 0, 0, 0], [0, 0, 3.5, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
        ]
    )
    symmetricTensor = ttb.tensor.from_data(symmetricData)
    assert symmetricTensor.issymmetric() is True
    assert symmetricTensor.issymmetric(version=1) is True
    answer, diffs, perms = symmetricTensor.issymmetric(version=1, return_details=True)
    assert answer is True
    assert (diffs == 0).all()

    symmetricData[3, 1, 0] = 3
    symmetricTensor = ttb.tensor.from_data(symmetricData)
    assert symmetricTensor.issymmetric() is False
    assert symmetricTensor.issymmetric(version=1) is False


@pytest.mark.indevelopment
def test_tensor_symmetrize(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way

    # Test new default version

    # 2-way
    symmetricData = np.array(
        [
            [[0.5, 0, 0.5, 0], [0, 0, 0, 0], [0.5, 0, 0, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 2.5, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
            [[0.5, 0, 0, 0], [0, 0, 0, 0], [0, 0, 3.5, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
        ]
    )
    symmetricTensor = ttb.tensor.from_data(symmetricData)
    assert symmetricTensor.symmetrize().isequal(symmetricTensor)

    # 3-way
    symmetricData = np.zeros((4, 4, 4))
    symmetricData[1, 2, 1] = 1
    symmetricTensor = ttb.tensor.from_data(symmetricData)
    print(f"\nsymmetricTensor:\n{symmetricTensor}")
    assert symmetricTensor.issymmetric() is False
    print(f"\nsymmetricTensor.symmetrize():\n{symmetricTensor.symmetrize()}")
    assert (symmetricTensor.symmetrize()).issymmetric()

    # 3-way
    shape = (2, 2, 2)
    T3 = ttb.tensor.from_data(np.arange(1, np.prod(shape) + 1), shape)
    T3sym = T3.symmetrize()
    print(f"\nT3sym:")
    print(T3sym)
    data3 = np.array(
        [
            [[1, 3 + 1 / 3], [3 + 1 / 3, 5 + 2 / 3]],
            [[3 + 1 / 3, 5 + 2 / 3], [5 + 2 / 3, 8]],
        ]
    )
    assert (T3sym.data == data3).all()

    # T3syms_2_1_3 = T3.symmetrize(grps=[[1], [0,2]])
    # print(f'\nT3syms_2_1_3:')
    # print(T3syms_2_1_3)

    with pytest.raises(AssertionError) as excinfo:
        symmetricTensor.symmetrize(grps=np.array([[0, 1], [1, 2]]))
    assert "Cannot have overlapping symmetries" in str(excinfo)

    # Improper shape tensor for symmetry
    asymmetricData = np.zeros((5, 4, 6))
    asymmetricTensor = ttb.tensor.from_data(asymmetricData)
    with pytest.raises(AssertionError) as excinfo:
        asymmetricTensor.symmetrize()
    assert "Dimension mismatch for symmetrization" in str(excinfo)

    # Test older keyword version
    symmetricData = np.array(
        [
            [[0.5, 0, 0.5, 0], [0, 0, 0, 0], [0.5, 0, 0, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 2.5, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
            [[0.5, 0, 0, 0], [0, 0, 0, 0], [0, 0, 3.5, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
        ]
    )

    symmetricTensor = ttb.tensor.from_data(symmetricData)
    assert symmetricTensor.symmetrize(version=1).isequal(symmetricTensor)

    symmetricData = np.zeros((4, 4, 4))
    symmetricData[1, 2, 1] = 1
    symmetricTensor = ttb.tensor.from_data(symmetricData)
    assert symmetricTensor.issymmetric() is False
    assert (symmetricTensor.symmetrize(version=1)).issymmetric()

    with pytest.raises(AssertionError) as excinfo:
        symmetricTensor.symmetrize(grps=np.array([[0, 1], [1, 2]]), version=1)
    assert "Cannot have overlapping symmetries" in str(excinfo)

    # Improper shape tensor for symmetry
    asymmetricData = np.zeros((5, 4, 6))
    asymmetricTensor = ttb.tensor.from_data(asymmetricData)
    with pytest.raises(AssertionError) as excinfo:
        asymmetricTensor.symmetrize(version=1)
    assert "Dimension mismatch for symmetrization" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor__str__(sample_tensor_2way):
    # Test 1D
    data = np.random.normal(size=(4,))
    tensorInstance = ttb.tensor.from_data(data)
    s = ""
    s += "tensor of shape "
    s += (" x ").join([str(int(d)) for d in tensorInstance.shape])
    s += "\n"
    s += "data"
    s += "[:] = \n"
    s += data.__str__()
    s += "\n"
    assert s == tensorInstance.__str__()

    # Test 2D
    data = np.random.normal(size=(4, 3))
    tensorInstance = ttb.tensor.from_data(data)
    s = ""
    s += "tensor of shape "
    s += (" x ").join([str(int(d)) for d in tensorInstance.shape])
    s += "\n"
    s += "data"
    s += "[:, :] = \n"
    s += data.__str__()
    s += "\n"
    assert s == tensorInstance.__str__()

    # Test 3D,shape in decreasing and increasing order
    data = np.random.normal(size=(4, 3, 2))
    tensorInstance = ttb.tensor.from_data(data)
    s = ""
    s += "tensor of shape "
    s += (" x ").join([str(int(d)) for d in tensorInstance.shape])
    s += "\n"
    for i in range(data.shape[0]):
        s += "data"
        s += "[{}, :, :] = \n".format(i)
        s += data[i, :, :].__str__()
        s += "\n"
    assert s == tensorInstance.__str__()

    data = np.random.normal(size=(2, 3, 4))
    tensorInstance = ttb.tensor.from_data(data)
    s = ""
    s += "tensor of shape "
    s += (" x ").join([str(int(d)) for d in tensorInstance.shape])
    s += "\n"
    for i in range(data.shape[0]):
        s += "data"
        s += "[{}, :, :] = \n".format(i)
        s += data[i, :, :].__str__()
        s += "\n"
    assert s == tensorInstance.__str__()

    # Test 4D
    data = np.random.normal(size=(4, 4, 3, 2))
    tensorInstance = ttb.tensor.from_data(data)
    s = ""
    s += "tensor of shape "
    s += (" x ").join([str(int(d)) for d in tensorInstance.shape])
    s += "\n"
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            s += "data"
            s += "[{}, {}, :, :] = \n".format(j, i)
            s += data[j, i, :, :].__str__()
            s += "\n"
    assert s == tensorInstance.__str__()

    # Test 5D
    data = np.random.normal(size=(2, 2, 2, 2, 2))
    tensorInstance = ttb.tensor.from_data(data)
    s = ""
    s += "tensor of shape "
    s += (" x ").join([str(int(d)) for d in tensorInstance.shape])
    s += "\n"
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            for k in range(data.shape[2]):
                s += "data"
                s += "[{}, {}, {}, :, :] = \n".format(k, j, i)
                s += data[k, j, i, :, :].__str__()
                s += "\n"
    assert s == tensorInstance.__str__()


@pytest.mark.indevelopment
def test_tensor_mttkrp(sample_tensor_2way):
    (params, tensorInstance) = sample_tensor_2way
    tensorInstance = ttb.tensor.from_function(np.ones, (2, 3, 4))

    # 2-way sparse tensor
    weights = np.array([2.0, 2.0])
    fm0 = np.array([[1.0, 3.0], [2.0, 4.0]])
    fm1 = np.array([[5.0, 8.0], [6.0, 9.0], [7.0, 10.0]])
    fm2 = np.array([[11.0, 15.0], [12.0, 16.0], [13.0, 17.0], [14.0, 18.0]])
    factor_matrices = [fm0, fm1, fm2]
    ktensorInstance = ttb.ktensor.from_data(weights, factor_matrices)

    m0 = np.array([[1800.0, 3564.0], [1800.0, 3564.0]])
    m1 = np.array([[300.0, 924.0], [300.0, 924.0], [300.0, 924.0]])
    m2 = np.array([[108.0, 378.0], [108.0, 378.0], [108.0, 378.0], [108.0, 378.0]])
    assert np.allclose(tensorInstance.mttkrp(ktensorInstance, 0), m0)
    assert np.allclose(tensorInstance.mttkrp(ktensorInstance, 1), m1)
    assert np.allclose(tensorInstance.mttkrp(ktensorInstance, 2), m2)

    # 5-way dense tensor
    shape = (2, 3, 4, 5, 6)
    T = ttb.tensor.from_data(np.arange(1, np.prod(shape) + 1), shape)
    U = []
    for s in shape:
        U.append(np.ones((s, 2)))

    data0 = np.array([[129600, 129600], [129960, 129960]])
    assert (T.mttkrp(U, 0) == data0).all()

    data1 = np.array([[86040, 86040], [86520, 86520], [87000, 87000]])
    assert (T.mttkrp(U, 1) == data1).all()

    data2 = np.array([[63270, 63270], [64350, 64350], [65430, 65430], [66510, 66510]])
    assert (T.mttkrp(U, 2) == data2).all()

    data3 = np.array(
        [[45000, 45000], [48456, 48456], [51912, 51912], [55368, 55368], [58824, 58824]]
    )
    assert (T.mttkrp(U, 3) == data3).all()

    data4 = np.array(
        [
            [7260, 7260],
            [21660, 21660],
            [36060, 36060],
            [50460, 50460],
            [64860, 64860],
            [79260, 79260],
        ]
    )
    assert (T.mttkrp(U, 4) == data4).all()

    # tensor too small
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance2 = ttb.tensor.from_data(np.array([1]))
        tensorInstance2.mttkrp([], 0)
    assert "MTTKRP is invalid for tensors with fewer than 2 dimensions" in str(excinfo)

    # second argument not a ktensor or list
    with pytest.raises(AssertionError) as excinfo:
        tensorInstance.mttkrp(5, 0)
    assert "Second argument should be a list of arrays or a ktensor" in str(excinfo)

    # second argument list is not the correct length
    with pytest.raises(AssertionError) as excinfo:
        m0 = np.ones((2, 2))
        tensorInstance.mttkrp([m0, m0, m0, m0], 0)
    assert "Second argument contains the wrong number of arrays" in str(excinfo)

    # arrays not the correct shape
    with pytest.raises(AssertionError) as excinfo:
        m0 = np.ones((2, 2))
        m1 = np.ones((3, 2))
        m2 = np.ones((5, 2))
        tensorInstance.mttkrp([m0, m1, m2], 0)
    assert "Entry 2 of list of arrays is wrong size" in str(excinfo)


@pytest.mark.indevelopment
def test_tensor_nvecs(sample_tensor_2way):
    (data, tensorInstance) = sample_tensor_2way

    nv1 = np.array([[0.4286671335486261, 0.5663069188480352, 0.7039467041474443]]).T
    nv2 = np.array(
        [
            [0.4286671335486261, 0.5663069188480352, 0.7039467041474443],
            [0.8059639085892916, 0.1123824140966059, -0.5811990803961161],
        ]
    ).T

    # Test for one eigenvector
    assert np.allclose((tensorInstance.nvecs(1, 1)), nv1)

    # Test for r >= N-1, requires cast to dense
    with pytest.warns(Warning) as record:
        assert np.allclose((tensorInstance.nvecs(1, 2)), nv2)
    assert (
        "Greater than or equal to tensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )



import pytest

import pyttb as ttb


def test_package_smoke():
    """A few sanity checks to make sure things don't explode"""
    assert len(ttb.__version__) > 0
    # Make sure warnings filter doesn't crash
    ttb.ignore_warnings(False)
    ttb.ignore_warnings(True)



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import os

import numpy as np

import pyttb as ttb

from .pyttb_utils import *


def import_data(filename):
    # Check if file exists
    if not os.path.isfile(filename):
        assert False, f"File path {filename} does not exist."

    # import
    fp = open(filename, "r")

    # tensor type should be on the first line
    # valid: tensor, sptensor, matrix, ktensor
    data_type = import_type(fp)

    if data_type not in ["tensor", "sptensor", "matrix", "ktensor"]:
        assert False, f"Invalid data type found: {data_type}"

    if data_type == "tensor":
        shape = import_shape(fp)
        data = import_array(fp, np.prod(shape))
        return ttb.tensor().from_data(data, shape)

    elif data_type == "sptensor":
        shape = import_shape(fp)
        nz = import_nnz(fp)
        subs, vals = import_sparse_array(fp, len(shape), nz)
        return ttb.sptensor().from_data(subs, vals, shape)

    elif data_type == "matrix":
        shape = import_shape(fp)
        mat = import_array(fp, np.prod(shape))
        mat = np.reshape(mat, np.array(shape))
        return mat

    elif data_type == "ktensor":
        shape = import_shape(fp)
        r = import_rank(fp)
        weights = import_array(fp, r)
        factor_matrices = []
        for n in range(len(shape)):
            fac_type = fp.readline().strip()
            fac_shape = import_shape(fp)
            fac = import_array(fp, np.prod(fac_shape))
            fac = np.reshape(fac, np.array(fac_shape))
            factor_matrices.append(fac)
        return ttb.ktensor().from_data(weights, factor_matrices)

    # Close file
    fp.close()


def import_type(fp):
    # Import IO data type
    return fp.readline().strip().split(" ")[0]


def import_shape(fp):
    # Import the shape of something from a file
    n = int(fp.readline().strip().split(" ")[0])
    shape = [int(d) for d in fp.readline().strip().split(" ")]
    if len(shape) != n:
        assert False, "Imported dimensions are not of expected size"
    return tuple(shape)


def import_nnz(fp):
    # Import the size of something from a file
    return int(fp.readline().strip().split(" ")[0])


def import_rank(fp):
    # Import the rank of something from a file
    return int(fp.readline().strip().split(" ")[0])


def import_sparse_array(fp, n, nz):
    # Import sparse data subs and vals from coordinate format data
    subs = np.zeros((nz, n), dtype="int64")
    vals = np.zeros((nz, 1))
    for k in range(nz):
        line = fp.readline().strip().split(" ")
        # 1-based indexing in file, 0-based indexing in package
        subs[k, :] = [np.int64(i) - 1 for i in line[:-1]]
        vals[k, 0] = line[-1]
    return subs, vals


def import_array(fp, n):
    return np.fromfile(fp, count=n, sep=" ")



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

__version__ = "1.6.0"

import warnings

from pyttb.cp_als import cp_als
from pyttb.cp_apr import *
from pyttb.export_data import export_data
from pyttb.hosvd import hosvd
from pyttb.import_data import import_data
from pyttb.khatrirao import khatrirao
from pyttb.ktensor import ktensor
from pyttb.pyttb_utils import *
from pyttb.sptenmat import sptenmat
from pyttb.sptensor import sptensor
from pyttb.sptensor3 import sptensor3
from pyttb.sumtensor import sumtensor
from pyttb.symktensor import symktensor
from pyttb.symtensor import symtensor
from pyttb.tenmat import tenmat
from pyttb.tensor import tensor
from pyttb.ttensor import ttensor
from pyttb.tucker_als import tucker_als


def ignore_warnings(ignore=True):
    if ignore:
        warnings.simplefilter("ignore")
    else:
        warnings.simplefilter("default")


ignore_warnings(True)



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np

import pyttb as ttb
from pyttb.pyttb_utils import *


def cp_als(
    tensor,
    rank,
    stoptol=1e-4,
    maxiters=1000,
    dimorder=None,
    init="random",
    printitn=1,
    fixsigns=True,
):
    """
    Compute CP decomposition with alternating least squares

    Parameters
    ----------
    tensor: :class:`pyttb.tensor` or :class:`pyttb.sptensor` or :class:`pyttb.ktensor`
    rank: int
        Rank of the decomposition
    stoptol: float
        Tolerance used for termination - when the change in the fitness function in successive iterations drops
        below this value, the iterations terminate (default: 1e-4)
    dimorder: list
        Order to loop through dimensions (default: [range(tensor.ndims)])
    maxiters: int
        Maximum number of iterations (default: 1000)
    init: str or :class:`pyttb.ktensor`
        Initial guess (default: "random")

             * "random": initialize using a :class:`pyttb.ktensor` with values chosen from a Normal distribution with mean 1 and standard deviation 0
             * "nvecs": initialize factor matrices of a :class:`pyttb.ktensor` using the eigenvectors of the outer product of the matricized input tensor
             * :class:`pyttb.ktensor`: initialize using a specific :class:`pyttb.ktensor` as input - must be the same shape as the input tensor and have the same rank as the input rank

    printitn: int
        Number of iterations to perform before printing iteration status - 0 for no status printing (default: 1)
    fixsigns: bool
        Align the signs of the columns of the factorization to align with the input tensor data (default: True)

    Returns
    -------
    M: :class:`pyttb.ktensor`
        Resulting ktensor from CP-ALS factorization
    Minit: :class:`pyttb.ktensor`
        Initial guess
    output: dict
        Information about the computation. Dictionary keys:

            * `params` : tuple of (stoptol, maxiters, printitn, dimorder)
            * `iters`: number of iterations performed
            * `normresidual`: norm of the difference between the input tensor and ktensor factorization
            * `fit`: value of the fitness function (fraction of tensor data explained by the model)

    Example
    -------
    Random initialization causes slight pertubation in intermediate results.
    `...` is our place holder for these numeric values.
    Example using default values ("random" initialization):

    >>> weights = np.array([1., 2.])
    >>> fm0 = np.array([[1., 2.], [3., 4.]])
    >>> fm1 = np.array([[5., 6.], [7., 8.]])
    >>> K = ttb.ktensor.from_data(weights, [fm0, fm1])
    >>> np.random.seed(1)
    >>> M, Minit, output = ttb.cp_als(K.full(), 2) # doctest: +ELLIPSIS
    CP_ALS:
     Iter 0: f = ... f-delta = ...
     Iter 1: f = ... f-delta = ...
     Final f = ...
    >>> print(M) # doctest: +ELLIPSIS
    ktensor of shape 2 x 2
    weights=[108.4715... 8.6114...]
    factor_matrices[0] =
    [[0.4187... 0.3989...]
     [0.9080... 0.9169...]]
    factor_matrices[1] =
    [[0.6188... 0.2581...]
     [0.7854... 0.9661...]]
    >>> print(Minit) # doctest: +ELLIPSIS
    ktensor of shape 2 x 2
    weights=[1. 1.]
    factor_matrices[0] =
    [[4.1702...e-01 7.2032...e-01]
     [1.1437...e-04 3.0233...e-01]]
    factor_matrices[1] =
    [[0.1467... 0.0923...]
     [0.1862... 0.3455...]]
    >>> print(output)
    {'params': (0.0001, 1000, 1, [0, 1]), 'iters': 1, 'normresidual': ..., 'fit': ...}

    Example using "nvecs" initialization:

    >>> M, Minit, output = ttb.cp_als(K.full(), 2, init="nvecs") # doctest: +ELLIPSIS
    CP_ALS:
     Iter 0: f = ... f-delta = ...
     Iter 1: f = ... f-delta = ...
     Final f = ...

    Example using :class:`pyttb.ktensor` initialization:

    >>> M, Minit, output = ttb.cp_als(K.full(), 2, init=K) # doctest: +ELLIPSIS
    CP_ALS:
     Iter 0: f = ... f-delta = ...
     Iter 1: f = ... f-delta = ...
     Final f = ...
    """

    # Extract number of dimensions and norm of tensor
    N = tensor.ndims
    normX = tensor.norm()

    # Set up dimorder if not specified
    if not dimorder:
        dimorder = list(range(N))
    else:
        if not isinstance(dimorder, list):
            assert False, "Dimorder must be a list"
        elif tuple(range(N)) != tuple(sorted(dimorder)):
            assert (
                False
            ), "Dimorder must be a list or permutation of range(tensor.ndims)"

    # Error checking
    assert rank > 0, "Number of components requested must be positive"

    # Set up and error checking on initial guess
    if isinstance(init, ttb.ktensor):
        # User provided an initial ktensor; validate it
        assert init.ndims == N, "Initial guess does not have {} modes".format(N)
        assert (
            init.ncomponents == rank
        ), "Initial guess does not have {} components".format(rank)
        for n in dimorder:
            if init.factor_matrices[n].shape != (tensor.shape[n], rank):
                assert False, "Mode {} of the initial guess is the wrong size".format(n)
    elif isinstance(init, str) and init.lower() == "random":
        factor_matrices = []
        for n in range(N):
            factor_matrices.append(np.random.uniform(0, 1, (tensor.shape[n], rank)))
        init = ttb.ktensor.from_factor_matrices(factor_matrices)
    elif isinstance(init, str) and init.lower() == "nvecs":
        factor_matrices = []
        for n in range(N):
            factor_matrices.append(tensor.nvecs(n, rank))
        init = ttb.ktensor.from_factor_matrices(factor_matrices)
    else:
        assert False, "The selected initialization method is not supported"

    # Set up for iterates and fit
    U = init.copy().factor_matrices
    fit = 0

    # Store the last MTTKRP result to accelerate fitness computation
    U_mttkrp = np.zeros((tensor.shape[dimorder[-1]], rank))

    if printitn > 0:
        print("CP_ALS:")

    # Main Loop: Iterate until convergence

    UtU = np.zeros((rank, rank, N))
    for n in range(N):
        UtU[:, :, n] = U[n].T @ U[n]

    for iter in range(maxiters):
        fitold = fit

        # Iterate over all N modes of the tensor
        for n in dimorder:
            # Calculate Unew = X_(n) * khatrirao(all U except n, 'r').
            Unew = tensor.mttkrp(U, n)

            # Save the last MTTKRP result for fitness check.
            if n == dimorder[-1]:
                U_mttkrp = Unew

            # Compute the matrix of coefficients for linear system
            Y = np.prod(UtU, axis=2, where=[i != n for i in range(N)])
            # don't try to solve linear system with Y = 0
            if (Y == np.zeros(Y.shape)).all():
                Unew = np.zeros(Unew.shape)
            else:
                Unew = np.linalg.solve(Y.T, Unew.T).T
            # TODO: should we have issparse implemented? I am not sure when the following will occur
            # if issparse(Unew):
            #    Unew = full(Unew)   # for the case R=1

            # Normalize each vector to prevent singularities in coefmatrix
            if iter == 0:
                weights = np.sqrt(sum(Unew**2, 0))  # 2-norm
            else:
                weights = np.maximum(np.max(np.abs(Unew), 0), 1)  # max-norm

            # if weights are 0, do not divide
            if not (weights == np.zeros(weights.shape)).all():
                Unew = Unew / weights

            U[n] = Unew
            UtU[:, :, n] = U[n].T @ U[n]

        M = ttb.ktensor.from_data(weights, U)

        # This is equivalent to innerprod(X,P).
        iprod = np.sum(
            np.sum(M.factor_matrices[dimorder[-1]] * U_mttkrp, 0) * weights, 0
        )
        if normX == 0:
            normresidual = M.norm() ** 2 - 2 * iprod
            fit = normresidual
        else:
            # the following input to np.sqrt can be negative due to rounding and truncation errors, so np.abs is used
            normresidual = np.sqrt(np.abs(normX**2 + M.norm() ** 2 - 2 * iprod))
            fit = 1 - (normresidual / normX)  # fraction explained by model

        fitchange = np.abs(fitold - fit)

        # Check for convergence
        if (iter > 0) and (fitchange < stoptol):
            flag = 0
        else:
            flag = 1

        if (divmod(iter, printitn)[1] == 0) or (printitn > 0 and flag == 0):
            print(f" Iter {iter}: f = {fit:e} f-delta = {fitchange:7.1e}")

        # Check for convergence
        if flag == 0:
            break

    # Clean up final result

    # Arrange the final tensor so that the columns are normalized.
    M.arrange()
    # Fix the signs if requested
    if fixsigns:
        M = M.fixsigns()

    if printitn > 0:
        if normX == 0:
            normresidual = M.norm() ** 2 - 2 * tensor.innerprod(M)
            fit = normresidual
        else:
            normresidual = np.sqrt(
                np.abs(normX**2 + M.norm() ** 2 - 2 * tensor.innerprod(M))
            )
            fit = 1 - (normresidual / normX)  # fraction explained by model
        print(f" Final f = {fit:e}")

    output = {}
    output["params"] = (stoptol, maxiters, printitn, dimorder)
    output["iters"] = iter
    output["normresidual"] = normresidual
    output["fit"] = fit

    return M, init, output


if __name__ == "__main__":
    import doctest  # pragma: no cover

    doctest.testmod()  # pragma: no cover



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np

import pyttb as ttb

from .pyttb_utils import *


class symktensor(object):
    """
    SYMKTENSOR Class for symmetric Kruskal tensors (decomposed).

    """

    def __init__(self, *args):  # pragma:no cover
        assert False, "SYMKTENSOR class not yet implemented"



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import os

import numpy as np

import pyttb as ttb

from .pyttb_utils import *


def export_data(data, filename, fmt_data=None, fmt_weights=None):
    """
    Export tensor-related data to a file.
    """
    # open file
    fp = open(filename, "w")

    if isinstance(data, ttb.tensor):
        print("tensor", file=fp)
        export_size(fp, data.shape)
        export_array(fp, data.data, fmt_data)

    elif isinstance(data, ttb.sptensor):
        print("sptensor", file=fp)
        export_sparse_size(fp, data)
        export_sparse_array(fp, data, fmt_data)

    elif isinstance(data, ttb.ktensor):
        print("ktensor", file=fp)
        export_size(fp, data.shape)
        export_rank(fp, data)
        export_weights(fp, data, fmt_weights)
        for n in range(data.ndims):
            print("matrix", file=fp)
            export_size(fp, data.factor_matrices[n].shape)
            export_factor(fp, data.factor_matrices[n], fmt_data)
        """
        fprintf(fid, 'ktensor\n');
        export_size(fid, size(A));
        export_rank(fid, A);
        export_lambda(fid, A.lambda, fmt_lambda);   
        for n = 1:length(size(A))
            fprintf(fid, 'matrix\n');
            export_size(fid, size(A.U{n}));
            export_factor(fid, A.U{n}, fmt_data);
        end
        """

    elif isinstance(data, np.ndarray):
        print("matrix", file=fp)
        export_size(fp, data.shape)
        export_array(fp, data, fmt_data)

    else:
        assert False, "Invalid data type for export"


def export_size(fp, shape):
    # Export the size of something to a file
    print(f"{len(shape)}", file=fp)  # # of dimensions on one line
    shape_str = " ".join([str(d) for d in shape])
    print(f"{shape_str}", file=fp)  # size of each dimensions on the next line


def export_rank(fp, data):
    # Export the rank of a ktensor to a file
    print(f"{len(data.weights)}", file=fp)  # ktensor rank on one line


def export_weights(fp, data, fmt_weights):
    # Export dense data that supports numel and linear indexing
    if not fmt_weights:
        fmt_weights = "%.16e"
    data.weights.tofile(fp, sep=" ", format=fmt_weights)
    print(file=fp)


def export_array(fp, data, fmt_data):
    # Export dense data that supports numel and linear indexing
    if not fmt_data:
        fmt_data = "%.16e"
    data.tofile(fp, sep="\n", format=fmt_data)
    print(file=fp)


def export_factor(fp, data, fmt_data):
    # Export dense data that supports numel and linear indexing
    if not fmt_data:
        fmt_data = "%.16e"
    for i in range(data.shape[0]):
        row = data[i, :]
        row.tofile(fp, sep=" ", format=fmt_data)
        print(file=fp)


def export_sparse_size(fp, A):
    # Export the size of something to a file
    print(f"{len(A.shape)}", file=fp)  # # of dimensions on one line
    shape_str = " ".join([str(d) for d in A.shape])
    print(f"{shape_str}", file=fp)  # size of each dimensions on the next line
    print(f"{A.nnz}", file=fp)  # number of nonzeros


def export_sparse_array(fp, A, fmt_data):
    # Export sparse array data in coordinate format
    if not fmt_data:
        fmt_data = "%.16e"
    # TODO: looping through all values may take a long time, can this be more efficient?
    for i in range(A.nnz):
        # 0-based indexing in package, 1-based indexing in file
        subs = A.subs[i, :] + 1
        subs.tofile(fp, sep=" ", format="%d")
        print(end=" ", file=fp)
        val = A.vals[i][0]
        val.tofile(fp, sep=" ", format=fmt_data)
        print(file=fp)



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import os

import numpy as np
import pytest

import pyttb as ttb


@pytest.fixture()
def sample_tensor():
    # truth data
    T = ttb.tensor.from_data(np.ones((3, 3, 3)), (3, 3, 3))
    return T


@pytest.fixture()
def sample_sptensor():
    # truth data
    subs = np.array(
        [
            [0, 0, 0],
            [0, 2, 2],
            [1, 1, 1],
            [1, 2, 0],
            [1, 2, 1],
            [1, 2, 2],
            [1, 3, 1],
            [2, 0, 0],
            [2, 0, 1],
            [2, 2, 0],
            [2, 2, 1],
            [2, 3, 0],
            [2, 3, 2],
            [3, 0, 0],
            [3, 0, 1],
            [3, 2, 0],
            [4, 0, 2],
            [4, 3, 2],
        ]
    )
    vals = np.reshape(np.array(range(1, 19)), (18, 1))
    shape = (5, 4, 3)
    S = ttb.sptensor().from_data(subs, vals, shape)
    return S


@pytest.fixture()
def sample_ktensor():
    # truth data
    weights = np.array([3, 2])
    fm0 = np.array([[1.0, 5.0], [2.0, 6.0], [3.0, 7.0], [4.0, 8.0]])
    fm1 = np.array([[2.0, 7.0], [3.0, 8.0], [4.0, 9.0], [5.0, 10.0], [6.0, 11.0]])
    fm2 = np.array([[3.0, 6.0], [4.0, 7.0], [5.0, 8.0]])
    factor_matrices = [fm0, fm1, fm2]
    K = ttb.ktensor.from_data(weights, factor_matrices)
    return K


@pytest.fixture()
def sample_array():
    # truth data
    M = np.array([[1.0, 5.0], [2.0, 6.0], [3.0, 7.0], [4.0, 8.0]])
    return M


@pytest.mark.indevelopment
def test_import_data_tensor(sample_tensor):
    # truth data
    T = sample_tensor

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "tensor.tns")
    X = ttb.import_data(data_filename)

    assert T.isequal(X)


@pytest.mark.indevelopment
def test_import_data_sptensor(sample_sptensor):
    # truth data
    S = sample_sptensor

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "sptensor.tns")
    X = ttb.import_data(data_filename)

    assert S.isequal(X)


@pytest.mark.indevelopment
def test_import_data_ktensor(sample_ktensor):
    # truth data
    K = sample_ktensor

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "ktensor.tns")
    X = ttb.import_data(data_filename)

    assert K.isequal(X)


@pytest.mark.indevelopment
def test_import_data_array(sample_array):
    # truth data
    M = sample_array

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "matrix.tns")
    X = ttb.import_data(data_filename)

    assert (M == X).all()


@pytest.mark.indevelopment
def test_export_data_tensor(sample_tensor):
    # truth data
    T = sample_tensor

    data_filename = os.path.join(os.path.dirname(__file__), "data", "tensor.out")
    ttb.export_data(T, data_filename)

    X = ttb.import_data(data_filename)
    assert T.isequal(X)
    os.unlink(data_filename)

    data_filename = os.path.join(os.path.dirname(__file__), "data", "tensor_int.out")
    ttb.export_data(T, data_filename, fmt_data="%d")

    X = ttb.import_data(data_filename)
    assert T.isequal(X)
    os.unlink(data_filename)


@pytest.mark.indevelopment
def test_export_data_sptensor(sample_sptensor):
    # truth data
    S = sample_sptensor

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "sptensor.out")
    ttb.export_data(S, data_filename)

    X = ttb.import_data(data_filename)
    assert S.isequal(X)
    os.unlink(data_filename)

    data_filename = os.path.join(os.path.dirname(__file__), "data", "sptensor_int.out")
    ttb.export_data(S, data_filename, fmt_data="%d")

    X = ttb.import_data(data_filename)
    assert S.isequal(X)
    os.unlink(data_filename)


@pytest.mark.indevelopment
def test_export_data_ktensor(sample_ktensor):
    # truth data
    K = sample_ktensor

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "ktensor.out")
    ttb.export_data(K, data_filename)

    X = ttb.import_data(data_filename)
    assert K.isequal(X)
    os.unlink(data_filename)

    data_filename = os.path.join(os.path.dirname(__file__), "data", "ktensor_int.out")
    ttb.export_data(K, data_filename, fmt_data="%d", fmt_weights="%d")

    X = ttb.import_data(data_filename)
    assert K.isequal(X)
    os.unlink(data_filename)


@pytest.mark.indevelopment
def test_export_data_array(sample_array):
    # truth data
    M = sample_array

    # imported data
    data_filename = os.path.join(os.path.dirname(__file__), "data", "matrix.out")
    ttb.export_data(M, data_filename)

    X = ttb.import_data(data_filename)
    assert (M == X).all()
    os.unlink(data_filename)

    data_filename = os.path.join(os.path.dirname(__file__), "data", "matrix_int.out")
    ttb.export_data(M, data_filename, fmt_data="%d")

    X = ttb.import_data(data_filename)
    assert (M == X).all()
    os.unlink(data_filename)



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np
import pytest

import pyttb as ttb

np.set_printoptions(precision=16)


@pytest.fixture()
def sample_ktensor_2way():
    weights = np.array([1.0, 2.0])
    fm0 = np.array([[1.0, 2.0], [3.0, 4.0]])
    fm1 = np.array([[5.0, 6.0], [7.0, 8.0]])
    factor_matrices = [fm0, fm1]
    data = {"weights": weights, "factor_matrices": factor_matrices}
    ktensorInstance = ttb.ktensor.from_data(weights, factor_matrices)
    return data, ktensorInstance


@pytest.fixture()
def sample_ktensor_3way():
    rank = 2
    shape = np.array([2, 3, 4])
    vector = np.arange(1, rank * sum(shape) + 1).astype(float)
    weights = 2 * np.ones(rank).astype(float)
    vector_with_weights = np.concatenate((weights, vector), axis=0)
    # vector_with_weights = vector_with_weights.reshape((len(vector_with_weights), 1))
    # ground truth
    fm0 = np.array([[1.0, 3.0], [2.0, 4.0]])
    fm1 = np.array([[5.0, 8.0], [6.0, 9.0], [7.0, 10.0]])
    fm2 = np.array([[11.0, 15.0], [12.0, 16.0], [13.0, 17.0], [14.0, 18.0]])
    factor_matrices = [fm0, fm1, fm2]
    data = {
        "weights": weights,
        "factor_matrices": factor_matrices,
        "vector": vector,
        "vector_with_weights": vector_with_weights,
        "shape": shape,
    }
    ktensorInstance = ttb.ktensor.from_data(weights, factor_matrices)
    return data, ktensorInstance


@pytest.fixture()
def sample_ktensor_symmetric():
    weights = np.array([1.0, 1.0])
    fm0 = np.array(
        [[2.340431417384394, 4.951967353890655], [4.596069112758807, 8.012451489774961]]
    )
    fm1 = np.array(
        [[2.340431417384394, 4.951967353890655], [4.596069112758807, 8.012451489774961]]
    )
    factor_matrices = [fm0, fm1]
    data = {"weights": weights, "factor_matrices": factor_matrices}
    ktensorInstance = ttb.ktensor.from_data(weights, factor_matrices)
    return data, ktensorInstance


@pytest.mark.indevelopment
def test_ktensor_init():
    empty = np.array([])

    # No args
    K0 = ttb.ktensor()
    assert (K0.weights == empty).all()
    assert K0.factor_matrices == []


@pytest.mark.indevelopment
def test_ktensor_from_tensor_type(sample_ktensor_2way):
    (data, K0) = sample_ktensor_2way
    K1 = ttb.ktensor.from_tensor_type(K0)
    assert (K0.weights == K1.weights).all()
    assert (K0.factor_matrices[0] == K1.factor_matrices[0]).all()
    assert (K0.factor_matrices[1] == K1.factor_matrices[1]).all()
    # won't work with instances other than ktensors
    with pytest.raises(AssertionError) as excinfo:
        K2 = ttb.ktensor.from_tensor_type(np.ones((2, 2)))
    assert "Cannot convert from <class 'numpy.ndarray'> to ktensor" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_from_factor_matrices(sample_ktensor_2way):
    (data, K0) = sample_ktensor_2way
    K0 = ttb.ktensor.from_factor_matrices(data["factor_matrices"])
    assert (K0.weights == np.ones(2)).all()
    assert (K0.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K0.factor_matrices[1] == data["factor_matrices"][1]).all()

    # Create ktensor with weights and multiple factor matrices as arguments
    K1 = ttb.ktensor.from_factor_matrices(
        data["factor_matrices"][0], data["factor_matrices"][1]
    )
    assert (K1.weights == np.ones(2)).all()
    assert (K1.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K1.factor_matrices[1] == data["factor_matrices"][1]).all()


@pytest.mark.indevelopment
def test_ktensor_from_data(sample_ktensor_2way, capsys):
    (data, K0) = sample_ktensor_2way
    K0 = ttb.ktensor.from_data(data["weights"], data["factor_matrices"])
    assert (K0.weights == data["weights"]).all()
    assert (K0.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K0.factor_matrices[1] == data["factor_matrices"][1]).all()

    # Create ktensor with weights and multiple factor matrices as arguments
    K1 = ttb.ktensor.from_data(
        data["weights"], data["factor_matrices"][0], data["factor_matrices"][1]
    )
    assert (K1.weights == data["weights"]).all()
    assert (K1.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K1.factor_matrices[1] == data["factor_matrices"][1]).all()

    # Weights that are int should be converted
    weights_int = np.array([1, 2])
    K2 = ttb.ktensor.from_data(weights_int, data["factor_matrices"])
    out, err = capsys.readouterr()
    assert (
        "converting weights from int64 to float" in out
        or "converting weights from int32 to float" in out
    )

    # Weights that are int should be converted
    fm0 = np.array([[1, 2], [3, 4]])
    fm1 = np.array([[5, 6], [7, 8]])
    factor_matrices = [fm0, fm1]
    K3 = ttb.ktensor.from_data(data["weights"], factor_matrices)
    out, err = capsys.readouterr()
    assert (
        "converting factor_matrices[0] from int64 to float" in out
        or "converting factor_matrices[0] from int32 to float" in out
    )


@pytest.mark.indevelopment
def test_ktensor_from_function():
    K0 = ttb.ktensor.from_function(np.ones, (2, 3, 4), 2)
    assert (K0.weights == np.array([1.0, 1.0])).all()
    assert (K0.factor_matrices[0] == np.ones((2, 2))).all()
    assert (K0.factor_matrices[1] == np.ones((3, 2))).all()
    assert (K0.factor_matrices[2] == np.ones((4, 2))).all()

    np.random.seed(1)
    K1 = ttb.ktensor.from_function(np.random.random_sample, (2, 3, 4), 2)
    assert (K1.weights == np.array([1.0, 1.0])).all()
    fm0 = np.array([[4.17022005e-01, 7.20324493e-01], [1.14374817e-04, 3.02332573e-01]])
    fm1 = np.array(
        [[0.14675589, 0.09233859], [0.18626021, 0.34556073], [0.39676747, 0.53881673]]
    )
    fm2 = np.array(
        [
            [0.41919451, 0.6852195],
            [0.20445225, 0.87811744],
            [0.02738759, 0.67046751],
            [0.4173048, 0.55868983],
        ]
    )
    assert np.linalg.norm(K1.factor_matrices[0] - fm0) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[1] - fm1) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[2] - fm2) < 1e-8


@pytest.mark.indevelopment
def test_ktensor_from_vector(sample_ktensor_3way):
    (data, K0) = sample_ktensor_3way

    # without explicit weights in x
    K0 = ttb.ktensor.from_vector(data["vector"], data["shape"], False)
    assert (K0.weights == np.ones((3, 1))).all()
    assert (K0.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K0.factor_matrices[1] == data["factor_matrices"][1]).all()
    assert (K0.factor_matrices[2] == data["factor_matrices"][2]).all()

    # with explicit weights in x
    K1 = ttb.ktensor.from_vector(data["vector_with_weights"], data["shape"], True)
    assert (K1.weights == data["weights"]).all()
    assert (K1.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K1.factor_matrices[1] == data["factor_matrices"][1]).all()
    assert (K1.factor_matrices[2] == data["factor_matrices"][2]).all()

    # data as a row vector will work, but will be transposed
    transposed_data = data["vector"].copy().reshape((1, len(data["vector"])))
    K2 = ttb.ktensor.from_vector(transposed_data, data["shape"], False)
    assert (K2.weights == np.ones((3, 1))).all()
    assert (K2.factor_matrices[0] == data["factor_matrices"][0]).all()
    assert (K2.factor_matrices[1] == data["factor_matrices"][1]).all()
    assert (K2.factor_matrices[2] == data["factor_matrices"][2]).all()

    # error if the shape does not match the the number of data elements
    with pytest.raises(AssertionError) as excinfo:
        K3 = ttb.ktensor.from_vector(data["vector"].T, data["shape"] + 7, False)
    assert "Input parameter 'data' is not the right length." in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_arrange(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way

    # permutation only
    K0 = ttb.ktensor.from_tensor_type(K)
    p = [1, 0]
    K0.arrange(permutation=p)
    assert (K0.weights == data["weights"][p]).all()
    assert (K0.factor_matrices[0] == data["factor_matrices"][0][:, p]).all()
    assert (K0.factor_matrices[1] == data["factor_matrices"][1][:, p]).all()

    # normalize and arrange by sorting weights (default)
    K1 = ttb.ktensor.from_tensor_type(K)
    K1.arrange()
    weights = np.array([89.4427191, 27.20294102])
    fm0 = np.array([[0.4472136, 0.31622777], [0.89442719, 0.9486833]])
    fm1 = np.array([[0.6, 0.58123819], [0.8, 0.81373347]])
    assert np.linalg.norm(K1.weights - weights) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[0] - fm0) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[1] - fm1) < 1e-8

    # error, cannot shoft weight and permute simultaneously
    with pytest.raises(AssertionError) as excinfo:
        K1.arrange(weight_factor=0, permutation=p)
    assert (
        "Weighting and permuting the ktensor at the same time is not allowed."
        in str(excinfo)
    )

    # error, length of permutation must equal number of components in ktensor
    with pytest.raises(AssertionError) as excinfo:
        K1.arrange(permutation=[0, 1, 2])
    assert (
        "Number of elements in permutation does not match number of components in ktensor."
        in str(excinfo)
    )


@pytest.mark.indevelopment
def test_ktensor_copy(sample_ktensor_2way):
    (data, K0) = sample_ktensor_2way
    K1 = K0.copy()
    assert (K0.weights == K1.weights).all()
    assert (K0.factor_matrices[0] == K1.factor_matrices[0]).all()
    assert (K0.factor_matrices[1] == K1.factor_matrices[1]).all()

    # make sure it is a deep copy
    K1.weights[0] = 0
    assert not (K0.weights[0] == K1.weights[0])


@pytest.mark.indevelopment
def test_ktensor_double(sample_ktensor_2way, sample_ktensor_3way):
    (data2, K2) = sample_ktensor_2way
    assert (K2.double() == np.array([[29.0, 39.0], [63.0, 85.0]])).all()
    (data3, K3) = sample_ktensor_3way
    A = np.array(
        [
            830.0,
            888.0,
            946.0,
            1004.0,
            942.0,
            1008.0,
            1074.0,
            1140.0,
            1054.0,
            1128.0,
            1202.0,
            1276.0,
            1180.0,
            1264.0,
            1348.0,
            1432.0,
            1344.0,
            1440.0,
            1536.0,
            1632.0,
            1508.0,
            1616.0,
            1724.0,
            1832.0,
        ]
    ).reshape((2, 3, 4))
    assert (K3.double() == A).all()


@pytest.mark.indevelopment
def test_ktensor_end(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way
    assert K.end() == 23
    assert K.end(k=0) == 1
    assert K.end(k=1) == 2
    assert K.end(k=2) == 3


@pytest.mark.indevelopment
def test_ktensor_extract(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way
    weights = data["weights"][[1]]
    fm0 = data["factor_matrices"][0][:, [1]]
    fm1 = data["factor_matrices"][1][:, [1]]
    fm2 = data["factor_matrices"][2][:, [1]]
    factor_matrices = [fm0, fm1, fm2]
    K_new = ttb.ktensor.from_data(weights, factor_matrices)

    # int
    K_extracted = K.extract(1)
    assert K_new.isequal(K_extracted)
    # tuple
    K_extracted = K.extract((1))
    assert K_new.isequal(K_extracted)
    # list
    K_extracted = K.extract([1])
    assert K_new.isequal(K_extracted)
    # np.ndarray
    K_extracted = K.extract(np.array([1]))
    assert K_new.isequal(K_extracted)

    # should return copy of ktensor
    K_extracted = K.extract()
    assert K.isequal(K_extracted)

    # wrong component index type
    with pytest.raises(AssertionError) as excinfo:
        K.extract(1.0)
    assert "Input parameter must be an int, tuple, list or numpy.ndarray" in str(
        excinfo
    )

    # too many components
    with pytest.raises(AssertionError) as excinfo:
        K.extract([0, 1, 2, 3])
    assert (
        "Number of components requested is not valid: 4 (should be in [1,...,2])."
        in str(excinfo)
    )

    # component index out of range
    with pytest.raises(AssertionError) as excinfo:
        K.extract((5))
    assert "Invalid component indices to be extracted: [5] not in range(2)" in str(
        excinfo
    )


@pytest.mark.indevelopment
def test_ktensor_fixsigns(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    K1 = ttb.ktensor.from_tensor_type(K)
    K2 = ttb.ktensor.from_tensor_type(K)

    # use same ktensor
    K1.factor_matrices[0][1, 1] = -K1.factor_matrices[0][1, 1]
    K1.factor_matrices[1][1, 1] = -K1.factor_matrices[1][1, 1]
    K2.factor_matrices[0][0, 1] = -K2.factor_matrices[0][0, 1]
    K2.factor_matrices[1][0, 1] = -K2.factor_matrices[1][0, 1]
    assert K1.fixsigns().isequal(K2)

    # use different ktensor for fixing the signs
    K3 = K.copy()
    K3.factor_matrices[0][1, 1] = -K3.factor_matrices[0][1, 1]
    K3.factor_matrices[1][1, 1] = -K3.factor_matrices[1][1, 1]
    K = K.fixsigns(K3)
    weights1 = np.array([27.202941017470888, 89.44271909999159])
    factor_matrix10 = np.array(
        [
            [0.3162277660168379, -0.4472135954999579],
            [0.9486832980505138, -0.8944271909999159],
        ]
    )
    factor_matrix11 = np.array([[0.5812381937190965, -0.6], [0.813733471206735, -0.8]])
    assert np.linalg.norm(K.weights - weights1) < 1e-8
    assert np.linalg.norm(K.factor_matrices[0] - factor_matrix10) < 1e-8
    assert np.linalg.norm(K.factor_matrices[1] - factor_matrix11) < 1e-8


@pytest.mark.indevelopment
def test_ktensor_full(sample_ktensor_2way, sample_ktensor_3way):
    (data, K2) = sample_ktensor_2way
    assert K2.full().isequal(
        ttb.tensor.from_data(np.array([[29.0, 39.0], [63.0, 85.0]]), (2, 2))
    )
    (data, K3) = sample_ktensor_3way
    print(K3.full())


@pytest.mark.indevelopment
def test_ktensor_innerprod(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    assert K.innerprod(K) == 13556

    # test with tensor
    Tdata = np.array([[1, 2], [3, 4]])
    Tshape = (2, 2)
    T = ttb.tensor().from_data(Tdata, Tshape)
    assert K.innerprod(T) == 636

    # test with sptensor
    Ssubs = np.array([[0, 0], [0, 1], [1, 1]])
    Svals = np.array([[0.5], [1.0], [1.5]])
    Sshape = (2, 2)
    S = ttb.sptensor().from_data(Ssubs, Svals, Sshape)
    assert K.innerprod(S) == 181

    # Wrong shape
    K1 = ttb.ktensor.from_function(np.ones, (2, 3), 2)
    with pytest.raises(AssertionError) as excinfo:
        K.innerprod(K1)
    assert "Innerprod can only be computed for tensors of the same size" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_isequal(sample_ktensor_2way):
    (data, K0) = sample_ktensor_2way
    # should be equal
    K1 = ttb.ktensor.from_tensor_type(K0)
    assert K0.isequal(K1)
    # ncomponents don't match
    K2 = ttb.ktensor.from_function(np.ones, (2, 2), 3)
    assert ~(K0.isequal(K2))
    # weights don't match
    K3 = ttb.ktensor.from_tensor_type(K0)
    K3.weights[0] = 10
    assert ~(K0.isequal(K3))
    # types don't match
    assert ~(K0.isequal(np.array([])))
    # factor_matrices don't match
    K4 = ttb.ktensor.from_tensor_type(K0)
    K4.factor_matrices[0] = np.zeros((2, 2))
    assert ~(K0.isequal(K4))


@pytest.mark.indevelopment
def test_ktensor_issymetric(sample_ktensor_2way, sample_ktensor_symmetric):
    # should not be symmetric
    (data, K) = sample_ktensor_2way
    assert ~(K.issymmetric())
    issym, diffs = K.issymmetric(return_diffs=True)
    assert (diffs == np.array([[0.0, 8.0], [0.0, 0]])).all()

    # should be symmetric
    (datas, K1) = sample_ktensor_symmetric
    assert K1.issymmetric()
    issym1, diffs1 = K1.issymmetric(return_diffs=True)
    assert (diffs1 == np.array([[0.0, 0.0], [0.0, 0]])).all()

    # Wrong shape
    K2 = ttb.ktensor.from_function(np.ones, (2, 3), 2)
    assert ~(K2.issymmetric())
    issym2, diffs2 = K2.issymmetric(return_diffs=True)
    assert (diffs2 == np.array([[0.0, np.inf], [0.0, 0]])).all()


@pytest.mark.indevelopment
def test_ktensor_mask(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    W = ttb.tensor.from_data(np.array([[0, 1], [1, 0]]))
    assert (K.mask(W) == np.array([[63], [39]])).all()

    # Mask too large
    with pytest.raises(AssertionError) as excinfo:
        K.mask(ttb.tensor.from_function(np.ones, (2, 3, 4)))
    assert "Mask cannot be bigger than the data tensor" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_mttkrp(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way
    K1 = ttb.ktensor.from_function(np.ones, (2, 3, 4), 4)
    output0 = np.array(
        [[12492.0, 12492.0, 12492.0, 12492.0], [17856.0, 17856.0, 17856.0, 17856.0]]
    )
    assert (K.mttkrp(K1.factor_matrices, 0) == output0).all()
    output1 = np.array(
        [
            [8892.0, 8892.0, 8892.0, 8892],
            [10116.0, 10116.0, 10116.0, 10116],
            [11340.0, 11340.0, 11340.0, 11340],
        ]
    )
    assert (K.mttkrp(K1.factor_matrices, 1) == output1).all()
    output2 = np.array(
        [
            [6858.0, 6858.0, 6858.0, 6858],
            [7344.0, 7344.0, 7344.0, 7344],
            [7830.0, 7830.0, 7830.0, 7830],
            [8316.0, 8316.0, 8316.0, 8316],
        ]
    )
    assert (K.mttkrp(K1.factor_matrices, 2) == output2).all()

    # Wrong number of factor matrices
    fm_wrong_size = [
        K1.factor_matrices[0],
        K1.factor_matrices[0],
        K1.factor_matrices[0],
        np.ones((5, 4)),
    ]
    with pytest.raises(AssertionError) as excinfo:
        K.mttkrp(fm_wrong_size, 0)
    assert "List of factor matrices is the wrong length" in str(excinfo)
    # Wrong input type
    fm_wrong_type = (
        K1.factor_matrices[0],
        K1.factor_matrices[0],
        K1.factor_matrices[0],
    )
    with pytest.raises(AssertionError) as excinfo:
        K.mttkrp(fm_wrong_type, 0)
    assert "Second argument must be list of numpy.ndarray's" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_ncomponents(sample_ktensor_2way):
    (data, K0) = sample_ktensor_2way
    assert K0.ncomponents == 2


@pytest.mark.indevelopment
def test_ktensor_ndims(sample_ktensor_2way, sample_ktensor_3way):
    (data, K0) = sample_ktensor_2way
    assert K0.ndims == 2
    data, K1 = sample_ktensor_3way
    assert K1.ndims == 3


@pytest.mark.indevelopment
def test_ktensor_norm():
    K0 = ttb.ktensor.from_function(np.zeros, (2, 3, 4), 2)
    assert pytest.approx(K0.norm(), 1e-8) == 0

    K1 = ttb.ktensor.from_function(np.ones, (2, 3, 4), 2)
    assert pytest.approx(K1.norm(), 1e-8) == 9.797958971132712

    rank = 2
    shape = np.array([2, 3, 4])
    data = np.arange(1, rank * sum(shape) + 1)
    weights = 2 * np.ones(rank)
    weights_and_data = np.concatenate((weights, data), axis=0)
    K2 = ttb.ktensor.from_vector(weights_and_data[:], shape, True)
    assert pytest.approx(K2.norm(), 1e-8) == 6.337788257744180e03


@pytest.mark.indevelopment
def test_ktensor_normalize(sample_ktensor_2way, sample_ktensor_3way):
    # get data and make several copies, so that all ktensor tests use the same data, as normalize computes in place
    data0, K0 = sample_ktensor_3way
    K1 = ttb.ktensor.from_tensor_type(K0)
    K2 = ttb.ktensor.from_tensor_type(K0)
    K3 = ttb.ktensor.from_tensor_type(K0)
    K4 = ttb.ktensor.from_tensor_type(K0)

    # normalize a single mode
    mode = 1
    # print("\nK0\n",K0)
    K0.normalize(mode=mode)
    # print("\nK0\n",K0)
    weights0 = np.array([20.97617696340303, 31.304951684997057])
    factor_matrix0 = np.array(
        [
            [0.4767312946227962, 0.5111012519999519],
            [0.5720775535473555, 0.5749889084999459],
            [0.6674238124719146, 0.6388765649999399],
        ]
    )
    assert np.linalg.norm(K0.weights - weights0) < 1e-8
    assert np.linalg.norm(K0.factor_matrices[mode] - factor_matrix0) < 1e-8

    # normalize using the defaults
    # print("\nK1\n",K1)
    K1.normalize()
    # print("\nK1\n",K1)
    weights1 = np.array([1177.285012220915, 5177.161384388167])
    factor_matrix10 = np.array([[0.4472135954999579, 0.6], [0.8944271909999159, 0.8]])
    factor_matrix11 = np.array(
        [
            [0.4767312946227962, 0.5111012519999519],
            [0.5720775535473555, 0.5749889084999459],
            [0.6674238124719146, 0.6388765649999399],
        ]
    )
    factor_matrix12 = np.array(
        [
            [0.4382504900892777, 0.4535055413676754],
            [0.4780914437337575, 0.4837392441255204],
            [0.5179323973782373, 0.5139729468833655],
            [0.5577733510227171, 0.5442066496412105],
        ]
    )
    assert np.linalg.norm(K1.weights - weights1) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[0] - factor_matrix10) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[1] - factor_matrix11) < 1e-8
    assert np.linalg.norm(K1.factor_matrices[2] - factor_matrix12) < 1e-8

    # normalize using vector 1-norm
    normtype = 1
    # print("\nK2\n",K2)
    K2.normalize(normtype=normtype)
    # print("\nK2\n",K2)
    weights2 = np.array([5400.0, 24948.0])
    factor_matrix20 = np.array(
        [
            [0.3333333333333333, 0.4285714285714285],
            [0.6666666666666666, 0.5714285714285714],
        ]
    )
    factor_matrix21 = np.array(
        [
            [0.2777777777777778, 0.2962962962962963],
            [0.3333333333333333, 0.3333333333333333],
            [0.3888888888888888, 0.3703703703703703],
        ]
    )
    factor_matrix22 = np.array(
        [
            [0.22, 0.2272727272727273],
            [0.24, 0.2424242424242424],
            [0.26, 0.2575757575757576],
            [0.28, 0.2727272727272727],
        ]
    )
    assert np.linalg.norm(K2.weights - weights2) < 1e-8
    assert np.linalg.norm(K2.factor_matrices[0] - factor_matrix20) < 1e-8
    assert np.linalg.norm(K2.factor_matrices[1] - factor_matrix21) < 1e-8
    assert np.linalg.norm(K2.factor_matrices[2] - factor_matrix22) < 1e-8

    # normalize and shift all weight to factor 1
    weight_factor = 1
    # print("\nK3\n",K3)
    K3.normalize(weight_factor=weight_factor)
    # print("\nK3\n",K3)
    weights3 = np.array([1.0, 1.0])
    factor_matrix30 = np.array(
        [[0.4472135954999579, 0.6000000000000001], [0.8944271909999159, 0.8]]
    )
    factor_matrix31 = np.array(
        [
            [561.2486080160912, 2646.0536653665963],
            [673.4983296193095, 2976.8103735374207],
            [785.7480512225277, 3307.567081708246],
        ]
    )
    factor_matrix32 = np.array(
        [
            [0.4382504900892776, 0.4535055413676753],
            [0.4780914437337574, 0.4837392441255204],
            [0.5179323973782373, 0.5139729468833654],
            [0.557773351022717, 0.5442066496412105],
        ]
    )
    assert np.linalg.norm(K3.weights - weights3) < 1e-8
    assert np.linalg.norm(K3.factor_matrices[0] - factor_matrix30) < 1e-8
    assert np.linalg.norm(K3.factor_matrices[1] - factor_matrix31) < 1e-8
    assert np.linalg.norm(K3.factor_matrices[2] - factor_matrix32) < 1e-8

    # error if the mode is not in the range of number of dimensions
    with pytest.raises(AssertionError) as excinfo:
        K0.normalize(mode=4)
    assert (
        "Parameter single_factor is invalid; index must be an int in range of number of dimensions"
        in str(excinfo)
    )

    # normalize and sort
    K4.normalize(sort=True)
    weights = np.array([5177.161384388167, 1177.285012220915])
    fm0 = np.array(
        [[0.6000000000000001, 0.4472135954999579], [0.8, 0.8944271909999159]]
    )
    fm1 = np.array(
        [
            [0.5111012519999519, 0.4767312946227962],
            [0.5749889084999459, 0.5720775535473555],
            [0.6388765649999399, 0.6674238124719146],
        ]
    )
    fm2 = np.array(
        [
            [0.4535055413676753, 0.4382504900892776],
            [0.4837392441255204, 0.4780914437337574],
            [0.5139729468833654, 0.5179323973782373],
            [0.5442066496412105, 0.557773351022717],
        ]
    )
    assert np.linalg.norm(K4.weights - weights) < 1e-8
    assert np.linalg.norm(K4.factor_matrices[0] - fm0) < 1e-8
    assert np.linalg.norm(K4.factor_matrices[1] - fm1) < 1e-8
    assert np.linalg.norm(K4.factor_matrices[2] - fm2) < 1e-8

    # distribute weight across all factors
    (data5, K5) = sample_ktensor_2way
    K5.normalize(weight_factor="all")
    weights = np.array([1.0, 1.0])
    fm0 = np.array(
        [
            [1.6493314105258194, 4.229485053762256],
            [4.947994231577458, 8.458970107524513],
        ]
    )
    fm1 = np.array(
        [[3.031531424242968, 5.674449654019056], [4.244143993940155, 7.565932872025407]]
    )
    assert np.allclose(K5.weights, weights)
    assert np.allclose(K5.factor_matrices[0], fm0)
    assert np.allclose(K5.factor_matrices[1], fm1)


@pytest.mark.indevelopment
def test_ktensor_nvecs(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way

    with pytest.warns(Warning) as record:
        assert np.allclose(
            K.nvecs(0, 1), np.array([[0.5731077440321353], [0.8194800264377384]])
        )
    assert (
        "Greater than or equal to ktensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )
    with pytest.warns(Warning) as record:
        assert np.allclose(
            K.nvecs(0, 2),
            np.array(
                [
                    [0.5731077440321353, 0.8194800264377384],
                    [0.8194800264377384, -0.5731077440321353],
                ]
            ),
        )
    assert (
        "Greater than or equal to ktensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )

    assert np.allclose(
        K.nvecs(1, 1),
        np.array([[0.5048631426517823], [0.5745404391632514], [0.6442177356747206]]),
    )
    with pytest.warns(Warning) as record:
        assert np.allclose(
            K.nvecs(1, 2),
            np.array(
                [
                    [0.5048631426517821, 0.7605567306550753],
                    [0.5745404391632517, 0.0568912743440822],
                    [0.6442177356747206, -0.6467741818894517],
                ]
            ),
        )
    assert (
        "Greater than or equal to ktensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )

    assert np.allclose(
        K.nvecs(2, 1),
        np.array(
            [
                [0.4507198734531968],
                [0.4827189140450413],
                [0.5147179546368857],
                [0.5467169952287301],
            ]
        ),
    )
    assert np.allclose(
        K.nvecs(2, 2),
        np.array(
            [
                [0.4507198734531969, 0.7048770074600103],
                [0.4827189140450412, 0.2588096791802433],
                [0.5147179546368857, -0.1872576491687805],
                [0.5467169952287302, -0.6333249775151949],
            ]
        ),
    )

    # Test for r >= N-1, requires cast to dense
    with pytest.warns(Warning) as record:
        K.nvecs(1, 3)
    assert (
        "Greater than or equal to ktensor.shape[n] - 1 eigenvectors requires cast to dense to solve"
        in str(record[0].message)
    )


@pytest.mark.indevelopment
def test_ktensor_permute(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way
    order = np.array([2, 0, 1])
    fm = [data["factor_matrices"][i] for i in order]
    K0 = ttb.ktensor.from_data(data["weights"], fm)
    assert K0.isequal(K.permute(order))

    # invalid permutation
    order_invalid = np.array([3, 0, 1])
    with pytest.raises(AssertionError) as excinfo:
        K.permute(order_invalid)
    assert "Invalid permutation" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_redistribute(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    K.redistribute(0)
    assert (np.array([[1, 4], [3, 8]]) == K[0]).all()
    assert (np.array([[5, 6], [7, 8]]) == K[1]).all()
    assert (np.array([1, 1]) == K.weights).all()


pytest.mark.indevelopment


def test_ktensor_score():
    A = ttb.ktensor.from_data(
        np.array([2, 1, 3]), np.ones((3, 3)), np.ones((4, 3)), np.ones((5, 3))
    )
    B = ttb.ktensor.from_data(
        np.array([2, 4]), np.ones((3, 2)), np.ones((4, 2)), np.ones((5, 2))
    )

    # defaults
    score, Aperm, flag, best_perm = A.score(B)
    assert score == 0.875
    assert np.allclose(Aperm.weights, np.array([15.49193338, 23.23790008, 7.74596669]))
    assert flag == 1
    assert (best_perm == np.array([0, 2, 1])).all()

    # compare just factor matrices (i.e., do not use weights)
    score, Aperm, flag, best_perm = A.score(B, weight_penalty=False)
    assert score == 1.0
    assert np.allclose(Aperm.weights, np.array([15.49193338, 7.74596669, 23.23790008]))
    assert flag == 1
    assert (best_perm == np.array([0, 1, 2])).all()

    # compute score using exhaustive search
    with pytest.raises(AssertionError) as excinfo:
        score, Aperm, flag, best_perm = A.score(B, greedy=False)
    assert "Not yet implemented. Only greedy method is implemented currently." in str(
        excinfo
    )

    # try to compute score with tensor type other than ktensor
    with pytest.raises(AssertionError) as excinfo:
        score, Aperm, flag, best_perm = A.score(ttb.tensor.from_tensor_type(B))
    assert "The first input should be a ktensor" in str(excinfo)

    # try to compute score when ktensor dimensions do not match
    with pytest.raises(AssertionError) as excinfo:
        # A is 3x4x5; B is 3x4x4
        B = ttb.ktensor.from_data(
            np.array([2, 4]), np.ones((3, 2)), np.ones((4, 2)), np.ones((4, 2))
        )
        score, Aperm, flag, best_perm = A.score(B)
    assert "Size mismatch" in str(excinfo)


pytest.mark.indevelopment


def test_ktensor_shape(sample_ktensor_2way, sample_ktensor_3way):
    (data, K0) = sample_ktensor_2way
    assert K0.shape == (2, 2)
    (data, K1) = sample_ktensor_3way
    assert K1.shape == (2, 3, 4)
    assert K1.shape[0] == 2
    assert K1.shape[1] == 3
    assert K1.shape[2] == 4
    K2 = ttb.ktensor()
    assert K2.shape == ()

    # error expected when parameter is not in range of number of dimensions
    with pytest.raises(IndexError) as excinfo:
        K1.shape[3]
    assert "tuple index out of range" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_symmetrize(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    K1 = K.symmetrize()
    weights = np.array([1.0, 1.0])
    fm = np.array(
        [[2.340431417384394, 4.951967353890656], [4.596069112758807, 8.01245148977496]]
    )
    assert np.allclose(K1.weights, weights)
    assert np.allclose(K1.factor_matrices[0], fm)
    assert np.allclose(K1.factor_matrices[1], fm)
    assert K1.issymmetric()

    # odd-ordered ktensor with negative weight
    weights = np.array([1.0, 2.0, -3.0])
    fm0 = np.ones((4, 3))
    fm1 = np.ones((4, 3))
    fm2 = -np.ones((4, 3))
    K2 = ttb.ktensor.from_data(weights, [fm0, fm1, fm2])
    K3 = K2.symmetrize()
    out_fm = np.array(
        [
            [-1.0, -1.2599210498948732, 1.442249570307408],
            [-1.0, -1.2599210498948732, 1.442249570307408],
            [-1.0, -1.2599210498948732, 1.442249570307408],
            [-1.0, -1.2599210498948732, 1.442249570307408],
        ]
    )
    assert np.allclose(K1.weights, np.ones((3, 1)))
    assert np.allclose(K3.factor_matrices[0], out_fm)
    assert np.allclose(K3.factor_matrices[1], out_fm)
    assert np.allclose(K3.factor_matrices[2], out_fm)
    assert K3.issymmetric()


@pytest.mark.indevelopment
def test_ktensor_tolist(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way

    # weights spread equally to all factor matrices
    fm0 = K.tolist()
    m0 = np.array(
        [
            [1.2599210498948732, 3.7797631496846193],
            [2.5198420997897464, 5.039684199579493],
        ]
    )
    m1 = np.array(
        [
            [6.299605249474366, 10.079368399158986],
            [7.559526299369239, 11.339289449053858],
            [8.819447349264113, 12.599210498948732],
        ]
    )
    m2 = np.array(
        [
            [13.859131548843605, 18.898815748423097],
            [15.119052598738477, 20.15873679831797],
            [16.37897364863335, 21.418657848212845],
            [17.638894698528226, 22.678578898107716],
        ]
    )
    assert np.allclose(fm0[0], m0)
    assert np.allclose(fm0[1], m1)
    assert np.allclose(fm0[2], m2)

    # weights are all 1
    K1 = ttb.ktensor.from_factor_matrices(fm0)
    fm0_tolist = K1.tolist()
    for i in range(len(fm0)):
        assert np.allclose(fm0[i], fm0_tolist[i])

    # weight spread to a single factor matrix
    fm1 = K.tolist(0)
    m0 = np.array(
        [
            [526.4978632435273, 3106.2968306329008],
            [1052.9957264870545, 4141.729107510534],
        ]
    )
    m1 = np.array(
        [
            [0.4767312946227962, 0.5111012519999519],
            [0.5720775535473555, 0.5749889084999459],
            [0.6674238124719146, 0.6388765649999399],
        ]
    )
    m2 = np.array(
        [
            [0.4382504900892776, 0.4535055413676753],
            [0.4780914437337574, 0.4837392441255204],
            [0.5179323973782373, 0.5139729468833654],
            [0.557773351022717, 0.5442066496412105],
        ]
    )
    assert np.allclose(fm1[0], m0)
    assert np.allclose(fm1[1], m1)
    assert np.allclose(fm1[2], m2)

    # mode not in range of self.ndims
    with pytest.raises(AssertionError) as excinfo:
        K.tolist(4)
    assert "Input parameter'mode' must be in the range of self.ndims" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor_tovec(sample_ktensor_3way):
    (data, K0) = sample_ktensor_3way
    assert (data["vector_with_weights"] == K0.tovec()).all()
    assert (data["vector"] == K0.tovec(include_weights=False)).all()


@pytest.mark.indevelopment
def test_ktensor_ttv(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way
    K0 = K.ttv(np.array([1, 1, 1]), dims=1)
    weights = np.array([36.0, 54.0])
    fm0 = np.array([[1.0, 3.0], [2.0, 4.0]])
    fm1 = np.array([[11.0, 15.0], [12.0, 16.0], [13.0, 17.0], [14.0, 18.0]])
    factor_matrices = [fm0, fm1]
    K1 = ttb.ktensor.from_data(weights, factor_matrices)
    assert K0.isequal(K1)

    # Empty dims requires that # vectors == # dimensions
    vec2 = np.array([1, 1])
    vec3 = np.array([1, 1, 1])
    vec4 = np.array([1, 1, 1, 1])
    assert K.ttv([vec2, vec3, vec4]) == 30348

    # Exclude dims should mirror dims
    assert K.ttv([vec2, vec3], dims=np.array([0, 1])).isequal(
        K.ttv([vec2, vec3], exclude_dims=2)
    )

    # Wrong shape
    with pytest.raises(AssertionError) as excinfo:
        K.ttv([vec2, vec3, np.array([1, 2])])
    assert "Multiplicand is wrong size" in str(excinfo)

    # Multiple dimensions, but fewer than all dimensions, not in same order as ktensor dimensions
    K2 = K.ttv([vec4, vec3], dims=np.array([2, 1]))
    weights = np.array([1800.0, 3564.0])
    fm0 = np.array([[1.0, 3.0], [2.0, 4.0]])
    assert K2.isequal(ttb.ktensor.from_data(weights, fm0))


@pytest.mark.indevelopment
def test_ktensor_update(sample_ktensor_3way):
    (data, K) = sample_ktensor_3way

    # single factor matrix updates
    K1 = ttb.ktensor.from_tensor_type(K)
    vec0 = np.random.randn(K.shape[0] * K.ncomponents)
    vec1 = np.random.randn(K.shape[1] * K.ncomponents)
    vec2 = np.random.randn(K.shape[2] * K.ncomponents)
    K1.update(0, vec0)
    assert (K1.factor_matrices[0] == vec0.reshape((K1.shape[0], K1.ncomponents))).all()
    K1.update(1, vec1)
    assert (K1.factor_matrices[1] == vec1.reshape((K1.shape[1], K1.ncomponents))).all()
    K1.update(2, vec2)
    assert (K1.factor_matrices[2] == vec2.reshape((K1.shape[2], K1.ncomponents))).all()

    # all factor matrix updates
    K2 = ttb.ktensor.from_tensor_type(K)
    vec_all = np.concatenate((vec0, vec1, vec2))
    K2.update([0, 1, 2], vec_all)
    assert (K2.factor_matrices[0] == vec0.reshape((K2.shape[0], K2.ncomponents))).all()
    assert (K2.factor_matrices[1] == vec1.reshape((K2.shape[1], K2.ncomponents))).all()
    assert (K2.factor_matrices[2] == vec2.reshape((K2.shape[2], K2.ncomponents))).all()

    # multiple but not all factor matrix updates
    K3 = ttb.ktensor.from_tensor_type(K)
    vec_some = np.concatenate((vec0, vec2))
    K3.update([0, 2], vec_some)
    assert (K3.factor_matrices[0] == vec0.reshape((K3.shape[0], K3.ncomponents))).all()
    assert (K3.factor_matrices[2] == vec2.reshape((K3.shape[2], K3.ncomponents))).all()

    # weights update
    weights = np.array([100, 200])
    K1.update(-1, weights)
    assert (K1.weights == weights).all()

    # not enough weights
    with pytest.raises(AssertionError) as excinfo:
        K1.update(-1, np.ones(1))
    assert "Data is too short" in str(excinfo)

    # not enough factor matrix values
    with pytest.raises(AssertionError) as excinfo:
        K1.update(0, np.ones(3))
    assert "Data is too short" in str(excinfo)

    # modes not sorted
    with pytest.raises(AssertionError) as excinfo:
        K1.update([0, 2, 1], np.ones(np.array(K.shape).prod()))
    assert "Modes must be sorted in ascending order" in str(excinfo)

    # invalid mode
    with pytest.raises(AssertionError) as excinfo:
        K1.update(4, np.array([]))
    assert "Invalid mode: 4" in str(excinfo)

    # too much data
    vec_all_plus = np.concatenate((vec_all, np.array([0.5])))
    with pytest.warns(Warning) as record:
        K2.update([0, 1, 2], vec_all_plus)
    assert "Failed to consume all of the input data" in str(record[0].message)


@pytest.mark.indevelopment
def test_ktensor__add__(sample_ktensor_2way, sample_ktensor_3way):
    (data0, K0) = sample_ktensor_2way
    (data1, K1) = sample_ktensor_3way
    # adding ktensor to itself
    K2 = K0 + K0
    assert (np.concatenate((data0["weights"], data0["weights"])) == K2.weights).all()
    for k in range(K2.ndims):
        assert (
            np.concatenate(
                (data0["factor_matrices"][k], data0["factor_matrices"][k]), axis=1
            )
            == K2.factor_matrices[k]
        ).all()

    # shapes do not match, should raise error
    with pytest.raises(AssertionError) as excinfo:
        K3 = K0 + K1
    assert "Must be two ktensors of the same shape" in str(excinfo)

    # types do not match, should raise error
    with pytest.raises(AssertionError) as excinfo:
        K0 + np.ones((2, 2))
    assert "Cannot add instance of this type to a ktensor" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        K0 + ttb.tensor()
    assert "Cannot add instance of this type to a ktensor" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        K0 + ttb.sptensor()
    assert "Cannot add instance of this type to a ktensor" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor__getitem__(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    # adding ktensor to itself
    assert K[0, 0] == 29
    assert K[0, 1] == 39
    assert K[1, 0] == 63
    assert K[1, 1] == 85
    assert (data["factor_matrices"][0] == K[0]).all()
    assert (data["factor_matrices"][1] == K[1]).all()
    # to return a 2D ndarray, the columns must be defined by a slice
    assert (data["factor_matrices"][0][:, [0]] == K[0][:, [0]]).all()
    with pytest.raises(AssertionError) as excinfo:
        K[0, 0, 0]
    assert "ktensor.__getitem__ requires tuples with 2 elements" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        K[5]
    assert (
        "ktensor.__getitem__() can only extract single elements (tuple of indices) or factor matrices (single index)"
        in str(excinfo)
    )


@pytest.mark.indevelopment
def test_ktensor__neg__(sample_ktensor_2way):
    (data0, K0) = sample_ktensor_2way
    # adding ktensor to itself
    K1 = -K0
    K2 = -K1
    assert (-data0["weights"] == K1.weights).all()
    for k in range(K1.ndims):
        assert (data0["factor_matrices"][k] == K1.factor_matrices[k]).all()
    assert K2.isequal(K0)


@pytest.mark.indevelopment
def test_ktensor__pos__(sample_ktensor_2way):
    (data0, K0) = sample_ktensor_2way
    # adding ktensor to itself
    K1 = +K0
    assert (data0["weights"] == K1.weights).all()
    for k in range(K1.ndims):
        assert (data0["factor_matrices"][k] == K1.factor_matrices[k]).all()
    assert K1.isequal(K0)


@pytest.mark.indevelopment
def test_ktensor__setitem__(sample_ktensor_2way):
    (data, K) = sample_ktensor_2way
    # adding ktensor to itself
    with pytest.raises(AssertionError) as excinfo:
        K[0, 0] = 1
    assert (
        "Subscripted assignment cannot be used to update individual elements of a ktensor."
        in str(excinfo)
    )


@pytest.mark.indevelopment
def test_ktensor__sub__(sample_ktensor_2way, sample_ktensor_3way):
    (data0, K0) = sample_ktensor_2way
    (data1, K1) = sample_ktensor_3way
    # adding ktensor to itself
    K2 = K0 - K0
    assert (np.concatenate((data0["weights"], -data0["weights"])) == K2.weights).all()
    for k in range(K2.ndims):
        assert (
            np.concatenate(
                (data0["factor_matrices"][k], data0["factor_matrices"][k]), axis=1
            )
            == K2.factor_matrices[k]
        ).all()
    # shapes do not match, should raise error
    with pytest.raises(AssertionError) as excinfo:
        K3 = K0 - K1
    assert "Must be two ktensors of the same shape" in str(excinfo)

    # types do not match, should raise error
    with pytest.raises(AssertionError) as excinfo:
        K0 - np.ones((2, 2))
    assert "Cannot subtract instance of this type from a ktensor" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        K0 - ttb.tensor()
    assert "Cannot subtract instance of this type from a ktensor" in str(excinfo)
    with pytest.raises(AssertionError) as excinfo:
        K0 - ttb.sptensor()
    assert "Cannot subtract instance of this type from a ktensor" in str(excinfo)


@pytest.mark.indevelopment
def test_ktensor__mul__(sample_ktensor_2way, sample_ktensor_3way):
    (data0, K0) = sample_ktensor_2way
    K1 = 2 * K0
    assert (2 * data0["weights"] == K1.weights).all()
    for k in range(K1.ndims):
        assert (data0["factor_matrices"][k] == K1.factor_matrices[k]).all()
    K2 = K0 * 2
    assert (2 * data0["weights"] == K2.weights).all()
    for k in range(K1.ndims):
        assert (data0["factor_matrices"][k] == K2.factor_matrices[k]).all()
    with pytest.raises(AssertionError) as excinfo:
        K3 = K0 * K0
    assert (
        "Multiplication by ktensors only allowed for scalars, tensors, or sptensors"
        in str(excinfo)
    )

    # test with tensor
    Tdata = np.array([[1, 2], [3, 4]])
    Tshape = (2, 2)
    T = ttb.tensor().from_data(Tdata, Tshape)
    K0T = K0 * T
    assert (K0T.double() == np.array([[29.0, 78.0], [189.0, 340.0]])).all()

    # test with sptensor
    Ssubs = np.array([[0, 0], [0, 1], [1, 1]])
    Svals = np.array([[0.5], [1.0], [1.5]])
    Sshape = (2, 2)
    S = ttb.sptensor().from_data(Ssubs, Svals, Sshape)
    K0S = S * K0
    assert (K0S.double() == np.array([[14.5, 39.0], [0.0, 127.5]])).all()


@pytest.mark.indevelopment
def test_ktensor__str__(sample_ktensor_2way):
    (data0, K0) = sample_ktensor_2way
    s = """ktensor of shape 2 x 2\nweights=[1. 2.]\nfactor_matrices[0] =\n[[1. 2.]\n [3. 4.]]\nfactor_matrices[1] =\n[[5. 6.]\n [7. 8.]]"""
    assert K0.__str__() == s



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np
import pytest

import pyttb as ttb


@pytest.mark.indevelopment
def test_khatrirao():
    A = np.array([[1, 2, 3], [4, 5, 6]])
    # This result was from MATLAB tensortoolbox, didn't verify by hand
    answer = np.array(
        [
            [1, 8, 27],
            [4, 20, 54],
            [4, 20, 54],
            [16, 50, 108],
            [4, 20, 54],
            [16, 50, 108],
            [16, 50, 108],
            [64, 125, 216],
        ]
    )
    assert (ttb.khatrirao([A, A, A]) == answer).all()
    assert (ttb.khatrirao([A, A, A], reverse=True) == answer).all()
    assert (ttb.khatrirao(A, A, A) == answer).all()

    # Test case where inputs are column vectors
    a_1 = np.array([[1], [1], [1], [1]])
    a_2 = np.array([[0], [1], [2], [3]])
    a_3 = np.array([[0, 0], [1, 0], [2, 0], [3, 0]])
    result = np.vstack(
        (
            a_2[0, 0] * np.ones((16, 1)),
            a_2[1, 0] * np.ones((16, 1)),
            a_2[2, 0] * np.ones((16, 1)),
            a_2[3, 0] * np.ones((16, 1)),
        )
    )
    assert (ttb.khatrirao([a_2, a_1, a_1]) == result).all()
    assert (ttb.khatrirao(a_2, a_1, a_1) == result).all()

    with pytest.raises(AssertionError) as excinfo:
        ttb.khatrirao([a_2, a_1, a_1], a_2)
    assert "Khatri Rao Acts on multiple Array arguments or a list of Arrays" in str(
        excinfo
    )

    with pytest.raises(AssertionError) as excinfo:
        ttb.khatrirao(a_2, a_1, np.ones((2, 2, 2)))
    assert "Each argument must be a matrix" in str(excinfo)

    with pytest.raises(AssertionError) as excinfo:
        ttb.khatrirao(a_2, a_1, a_3)
    assert "All matrices must have the same number of columns." in str(excinfo)
